#!/usr/bin/env python

# https://github.com/petraszd/postgres-db-diff/pull/1

# If you are reading this code and thinking: why this file have not been
# split into smaller and easier to read modules? The answer is quite simple:
# I want users to be able just copy/paste this file and run it
import argparse
import difflib
import os.path
import subprocess
import sys


def check_database_name(name):
    try:
        out = db_out(name, "SELECT 42", stderr=None)
    except subprocess.CalledProcessError:
        raise argparse.ArgumentTypeError(
            'Can not access DB using psql. Probably it does not exists.'
        )

    if '42' not in out:
        raise argparse.ArgumentTypeError(
            'Unknown problem executing SQL statements using psql. Aborting.'
        )

    return name


def check_diff_directory(name):
    path = os.path.join(name)
    if not os.path.exists(path):
        return name

    if not os.path.isdir(path):
        raise argparse.ArgumentTypeError('It is not a directory')

    if os.listdir(path):
        raise argparse.ArgumentTypeError('Directory must be empty')

    return name


def parser_arguments():
    parser = argparse.ArgumentParser()

    parser.add_argument('--db1', help='First DB name',
                        type=check_database_name, required=True)
    parser.add_argument('--db2', help='Second DB name',
                        type=check_database_name, required=True)
    parser.add_argument('--diff-folder',
                        help='Directory to output diffs',
                        type=check_diff_directory, required=False)
    parser.add_argument('--rowcount',
                        help='Compare tables row count',
                        action='store_true')

    return parser.parse_args()


def db_out(db_name, cmd, stderr=subprocess.STDOUT):
    return subprocess.check_output(
        "psql -d '{}' -c '{}'".format(db_name, cmd), shell=True, stderr=stderr
    ).decode('utf-8')


def get_table_rowcount(db_name, table_name, stderr=subprocess.STDOUT):
    cmd = 'select count(1) from "{}";'.format(table_name)
    output = subprocess.check_output(
        "psql -d '{}' -c '{}' --quiet --tuples-only".format(db_name, cmd), shell=True, stderr=stderr
    ).decode('utf-8')
    return int(output.strip())


def get_db_tables(db_name):
    tables = set()
    for line in db_out(db_name, '\\dt').splitlines():
        elems = line.split()
        if line and elems[0] == 'public':
            tables.add(elems[2])
    return tables


def get_db_views(db_name):
    views = set()
    for line in db_out(db_name, '\\dv').splitlines():
        elems = line.split()
        if line and elems[0] == 'public':
            views.add(elems[2])
    return views


def get_table_definition(db_name, table_name):
    lines = db_out(db_name, '\\d "{}"'.format(table_name)).splitlines()
    lines = [x for x in lines if x.strip()]

    columns_range = [None, None]
    indexes_range = [None, None]
    check_constr_range = [None, None]
    foreign_constr_range = [None, None]
    process_constr_range = [None, None]

    S_START = 1
    S_COLUMNS = 2
    S_INDEXES = 3
    S_CHECK_CONSTR = 4
    S_FOREIGN_CONSTR = 5
    S_REFERENCES = 6
    S_END = 7

    def replace_with_sorted(lines, a, b):
        if a is None or b is None:
            return lines
        return lines[:a] + sorted(lines[a:b]) + lines[b:]

    def get_after_columns_state(x):
        if x == 'Indexes:':
            return S_INDEXES
        elif x == 'Check constraints:':
            return S_CHECK_CONSTR
        elif x == 'Foreign-key constraints:':
            return S_FOREIGN_CONSTR
        elif x == 'Referenced by:':
            return S_REFERENCES
        return S_END

    def update_range(line_range, i):
        if line_range[0] is None:
            line_range[0] = i
            line_range[1] = i + 1
        else:
            line_range[1] = i + 1

    def process_start(i, x):
        if x[0:2] == '--':
            return S_COLUMNS
        return S_START

    def process_columns(i, x):
        if x[0] != ' ':
            return get_after_columns_state(x)
        update_range(columns_range, i)
        return S_COLUMNS

    def process_indexes(i, x):
        if x[0] != ' ':
            return get_after_columns_state(x)
        update_range(indexes_range, i)
        return S_INDEXES

    def process_check_constr(i, x):
        if x[0] != ' ':
            return get_after_columns_state(x)
        update_range(check_constr_range, i)
        return S_CHECK_CONSTR

    def process_foreign_constr(i, x):
        if x[0] != ' ':
            return get_after_columns_state(x)
        update_range(foreign_constr_range, i)
        return S_FOREIGN_CONSTR

    def process_references(i, x):
        if x[0] != ' ':
            return get_after_columns_state(x)
        update_range(process_constr_range, i)
        return S_REFERENCES

    def process_end(i, x):
        return S_END

    processes = {
        S_START: process_start,
        S_COLUMNS: process_columns,
        S_INDEXES: process_indexes,
        S_CHECK_CONSTR: process_check_constr,
        S_FOREIGN_CONSTR: process_foreign_constr,
        S_REFERENCES: process_references,
        S_END: process_end,
    }

    state = S_START
    for i, x in enumerate(lines):
        state = processes[state](i, x)

    lines = replace_with_sorted(lines, *columns_range)
    lines = replace_with_sorted(lines, *indexes_range)
    lines = replace_with_sorted(lines, *check_constr_range)
    lines = replace_with_sorted(lines, *foreign_constr_range)
    lines = replace_with_sorted(lines, *process_constr_range)
    return '\n'.join(lines)


def compare_number_of_items(options, db1_items, db2_items, items_name):
    if db1_items != db2_items:
        additional_db1 = db1_items - db2_items
        additional_db2 = db2_items - db1_items

        if additional_db1:
            sys.stdout.write(
                '{}: additional in "{}"\n'.format(items_name, options.db1)
            )
            for t in additional_db1:
                sys.stdout.write('\t{}\n'.format(t))
            sys.stdout.write('\n')

        if additional_db2:
            sys.stdout.write(
                '{}: additional in "{}"\n'.format(items_name, options.db2)
            )
            for t in additional_db2:
                sys.stdout.write('\t{}\n'.format(t))
            sys.stdout.write('\n')


# TODO: Using same function to compare tables and views. It is not very suited
# for views. But I do not see any clear way to have cleaner interface
def compare_each_table(options, db1_tables, db2_tables, items_name):
    not_matching_tables = []
    not_matching_rowcount = []

    for t in sorted(db1_tables & db2_tables):
        t1 = get_table_definition(options.db1, t)
        t2 = get_table_definition(options.db2, t)
        if t1 != t2:
            not_matching_tables.append(t)

            diff = difflib.unified_diff(
                [x + '\n' for x in t1.splitlines()],
                [x + '\n' for x in t2.splitlines()],
                '{}.{}.{}'.format(items_name, options.db1, t),
                '{}.{}.{}'.format(items_name, options.db2, t),
                n=sys.maxsize
            )

            if options.diff_folder:
                if not os.path.exists(options.diff_folder):
                    os.mkdir(options.diff_folder)
                filepath = os.path.join(
                    options.diff_folder, '{}.diff'.format(t)
                )
                with open(filepath, 'w') as f:
                    for diff_line in diff:
                        f.write(diff_line)

        elif options.rowcount:
            t1_rowcount = get_table_rowcount(options.db1, t)
            t2_rowcount = get_table_rowcount(options.db2, t)
            if t1_rowcount != t2_rowcount:
                not_matching_rowcount.append('{} ({} != {})'.format(t, t1_rowcount, t2_rowcount))

    if not_matching_tables:
        sys.stdout.write('{}: not matching\n'.format(items_name))
        for t in not_matching_tables:
            sys.stdout.write('\t{}\n'.format(t))
        sys.stdout.write('\n')

    if not_matching_rowcount:
        sys.stdout.write('{}: not matching rowcount\n'.format(items_name))
        for t in not_matching_rowcount:
            sys.stdout.write('\t{}\n'.format(t))
        sys.stdout.write('\n')


def main():
    options = parser_arguments()

    db1_tables = get_db_tables(options.db1)
    db2_tables = get_db_tables(options.db2)

    compare_number_of_items(options, db1_tables, db2_tables, 'TABLES')
    compare_each_table(options, db1_tables, db2_tables, 'TABLES')

    db1_views = get_db_views(options.db1)
    db2_views = get_db_views(options.db2)
    compare_number_of_items(options, db1_views, db2_views, 'VIEWS')
    compare_each_table(options, db1_views, db2_views, 'VIEWS')


if __name__ == "__main__":
    main()


import traceback
import subprocess
import base64
import yaml
import json
from ckan_cloud_operator import kubectl


FLAGS = [
    'forceCreateAnnotations',  # Set created annotations even if a component failed to create
    'skipDbPermissions', # Skip setting DB permissions on update
    'skipDatastorePermissions', # Skip setting Datastore permissions on update
]


STATUSES = {
    'db': ['created'],
    'datastore': ['created'],
    'deployment': ['created'],
    'envvars': ['created'],
    'instance': ['created'],
    'namespace': ['created'],
    'registry': ['created'],
    'solr': ['created'],
    'storage': ['created'],
    'ckan': ['created'],
}


# flexible annotations encoded to json and permitted using key prefixes
JSON_ANNOTATION_PREFIXES = [
    'router-traefik'
]


SECRET_ANNOTATIONS = [
    'databasePassword',
    'datastorePassword',
    'datastoreReadonlyUser',
    'datatastoreReadonlyPassword',
]


__NONE__ = object()


class DeisCkanInstanceAnnotations(object):
    """Manage annotations related to the instance which are used to store instance metadata"""

    def __init__(self, instance, override_flags=None, persist_overrides=False):
        self.instance = instance
        self._override_flags = override_flags
        if self._override_flags:
            for flag in self._override_flags:
                assert flag in FLAGS, f'invalid flag: {flag}'
            if persist_overrides and len(self._override_flags) > 0:
                self._annotate(*['{}={}'.format(k, 'true' if v else 'false')
                                 for k, v in self._override_flags.items()], overwrite=True)

    def set_status(self, key, status):
        assert status in STATUSES[key], f'invalid status: {key}={status}'
        self._annotate(f'{key}-{status}=true')

    def get_status(self, key, status):
        assert key in STATUSES.keys(), f'invalid status key: {key}'
        return bool(self._get_annotation(f'{key}-{status}'))

    def update_status(self, key, status, update_func, force_update=False):
        if self.get_status(key, status):
            if force_update:
                update_func()
                return True
            else:
                return False
        else:
            try:
                update_func()
            except:
                if self.get_flag('forceCreateAnnotations'):
                    traceback.print_exc()
                else:
                    raise
            self.set_status(key, status)
            return True

    def set_flag(self, flag):
        self._annotate(f'{flag}=true')

    def set_flags(self, *flags):
        self._annotate(*[f'{flag}=true' for flag in flags])

    def get_flag(self, flag):
        """flags  are boolean fields which can be added manually to modify operator functionality

        Flags can be set kubectl annotate on the DeisCkanInstance object, for example:

        kubectl -n ckan-cloud annotate DeisCkanInstance/atea32 ckan-cloud/forceCreateAnnotations=true
        """
        assert flag in FLAGS, f'invalid flag: {flag}'
        if self._override_flags and flag in self._override_flags:
            return self._override_flags[flag]
        else:
            value = self._get_annotation(flag)
            if type(value) == str:
                return False if value.lower().strip() in ['false', '0', 'no', ''] else bool(value)
            else:
                return bool(value)

    def get(self):
        data = {k.replace('ckan-cloud/', ''): v for k, v in self.instance.values['metadata'].get('annotations', {}).items()
                if k.startswith('ckan-cloud/')}
        data['ready'] = True
        return data

    def set_secrets(self, key_values):
        for key in key_values:
            assert key in SECRET_ANNOTATIONS, f'invalid secret key: {key}'
        secret = getattr(self, '_secret', None)
        cur_data = secret.get('data', {}) if secret and secret != __NONE__ else {}
        secret = kubectl.get(f'secret {self.instance.id}-annotations', namespace=self.instance.id, required=False)
        if not secret:
            secret = {'data': {}}
        secret['data'].update(**cur_data)
        for key, value in key_values.items():
            secret['data'][key] = base64.b64encode(value.encode()).decode()
        secret = {
            'apiVersion': 'v1',
            'kind': 'Secret',
            'metadata': {
                'name': f'{self.instance.id}-annotations',
                'namespace': self.instance.id
            },
            'type': 'Opaque',
            'data': secret['data']
        }
        subprocess.run(f'kubectl -n {self.instance.id} apply -f -',
                       input=yaml.dump(secret).encode(), shell=True, check=True)
        self._secret = secret

    def set_secret(self, key, value):
        self.set_secrets({key: value})

    def get_secret(self, key, default=None):
        assert key in SECRET_ANNOTATIONS, f'invalid secret key: {key}'
        secret = getattr(self, '_secret', None)
        if not secret:
            secret = kubectl.get(f'secret {self.instance.id}-annotations',
                                 namespace=self.instance.id, required=False)
            if not secret:
                secret = __NONE__
            self._secret = secret
        if secret and secret != __NONE__:
            value = secret.get('data', {}).get(key, None)
            return base64.b64decode(value).decode() if value else default
        else:
            return default

    def json_annotate(self, key, value, overwrite=True):
        ans = []
        assert any([key.startswith(prefix) for prefix in JSON_ANNOTATION_PREFIXES]), f'invalid json annotation key: {key}'
        value = json.dumps(value)
        ans.append(f"{key}='{value}'")
        self._annotate(*ans, overwrite=overwrite)

    def _annotate(self, *annotations, overwrite=True):
        cmd = f'kubectl -n ckan-cloud annotate {self.instance.kind} {self.instance.id}'
        for annotation in annotations:
            cmd += f' ckan-cloud/{annotation}'
        if overwrite:
            cmd += ' --overwrite'
        subprocess.check_call(cmd, shell=True)

    def _get_annotation(self, annotation, default=None):
        return self.instance.values['metadata'].get('annotations', {}).get(f'ckan-cloud/{annotation}', default)



import yaml
import hashlib

from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs
from ckan_cloud_operator.routers.annotations import CkanRoutersAnnotations
from ckan_cloud_operator.routers.traefik import manager as traefik_manager
from ckan_cloud_operator.routers.routes import manager as routes_manager
from ckan_cloud_operator.providers.routers.manager import get_env_id, get_default_root_domain, get_dns_provider


ROUTER_TYPES = {
    'traefik': {
        'default': True,
        'manager': traefik_manager
    }
}

DEFAULT_ROUTER_TYPE = [k for k,v in ROUTER_TYPES.items() if v.get('default')][0]


def create(router_name, router_spec):
    from ckan_cloud_operator.providers.cluster.manager import get_or_create_multi_user_volume_claim
    from ckan_cloud_operator.routers.traefik.deployment import get_label_suffixes

    router_type = router_spec.get('type')
    default_root_domain = router_spec.get('default-root-domain')
    assert router_type in ROUTER_TYPES and default_root_domain, f'Invalid router spec: {router_spec}'
    get(router_name, only_dns=True, failfast=True)
    print(f'Creating CkanCloudRouter {router_name} {router_spec}')
    labels = _get_labels(router_name, router_type)
    router = kubectl.get_resource('stable.viderum.com/v1', 'CkanCloudRouter', router_name, labels,
                                  spec=dict(router_spec, **{'type': router_type}))
    router_manager = ROUTER_TYPES[router_type]['manager']
    router = router_manager.create(router)
    get_or_create_multi_user_volume_claim(get_label_suffixes(router_name, router_type))
    annotations = CkanRoutersAnnotations(router_name, router)
    annotations.json_annotate('default-root-domain', default_root_domain)


def get_traefik_router_spec(default_root_domain=None, cloudflare_email=None, cloudflare_api_key=None,
                            wildcard_ssl_domain=None, external_domains=False, dns_provider=None):
    if not default_root_domain: default_root_domain = 'default'
    if not cloudflare_email: cloudflare_email = 'default'
    if not cloudflare_api_key: cloudflare_api_key = 'default'
    return {
        'type': 'traefik',
        'default-root-domain': default_root_domain,
        # the cloudflare spec is not saved as part of the CkanCloudRouter spec
        # it is removed and saved as a secret by the traefik router manager
        'cloudflare': {
            'email': cloudflare_email,
            'api-key': cloudflare_api_key
        },
        'wildcard-ssl-domain': wildcard_ssl_domain,
        'external-domains': bool(external_domains),
        'dns-provider': dns_provider
    }


def update(router_name, wait_ready=False, dry_run=False):
    router, spec, router_type, annotations, labels, router_type_config = _init_router(router_name)
    print(f'Updating CkanCloudRouter {router_name} (type={router_type}) (labels={labels})')
    routes = routes_manager.list(labels)
    router_type_config['manager'].update(router_name, wait_ready, spec, annotations, routes, dry_run=dry_run)


def list(full=False, values_only=False, async_print=True):
    res = None if async_print else []
    for router in kubectl.get('CkanCloudRouter')['items']:
        if values_only:
            data = {'name': router['metadata']['name'],
                    'type': router['spec']['type']}
        else:
            data = get(router)
            if not full:
                data = {'name': data['name'],
                        'type': data['type'],
                        'ready': data['ready']}
        if res is None:
            print(yaml.dump([data], default_flow_style=False))
        else:
            res.append(data)
    if res is not None:
        return res


def get(router_name_or_values, required=False, only_dns=False, failfast=False):
    if type(router_name_or_values) == str:
        router_name = router_name_or_values
        router_values = kubectl.get(f'CkanCloudRouter {router_name}', required=required)
    else:
        router_name = router_name_or_values['metadata']['name']
        router_values = router_name_or_values
    router, spec, router_type, annotations, labels, router_type_config = _init_router(router_name, router_values, required=required)
    if router:
        dns_data = router_type_config['manager'].get(router_name, 'dns', router, failfast=True)
        if not only_dns:
            deployment_data = router_type_config['manager'].get(router_name, 'deployment')
            routes = routes_manager.list(_get_labels(router_name, router_type))
        else:
            deployment_data = None
            routes = None
        if only_dns:
            return {'name': router_name,
                    'dns': dns_data}
        else:
            return {'name': router_name,
                    'annotations': router_values['metadata']['annotations'],
                    'routes': [route.get('spec') for route in routes] if routes else [],
                    'type': router_type,
                    'deployment': deployment_data,
                    'ready': deployment_data.get('ready', False),
                    'dns': dns_data,
                    'spec': {'ready': True, **router_values['spec']}}
    else:
        return None


def create_subdomain_route(router_name, route_spec, dry_run=False):
    target_type = route_spec['target-type']
    sub_domain = route_spec.get('sub-domain')
    root_domain = route_spec.get('root-domain')
    if target_type == 'datapusher':
        target_resource_id = route_spec['datapusher-name']
    elif target_type == 'deis-instance':
        target_resource_id = route_spec['deis-instance-id']
    elif target_type == 'backend-url':
        target_resource_id = route_spec['target-resource-id']
    elif target_type == 'ckan-instance':
        target_resource_id = route_spec['ckan-instance-id']
    elif target_type == 'app-instance':
        target_resource_id = route_spec['app-instance-id']
    else:
        raise Exception(f'Invalid route spec: {route_spec}')
    sub_domain, root_domain = _get_default_sub_root_domain(sub_domain, root_domain, target_resource_id)
    route_name = 'cc' + hashlib.sha3_224(f'{target_type} {target_resource_id} {root_domain} {sub_domain}'.encode()).hexdigest()

    logs.info(f'initializing router {router_name}')
    router, spec, router_type, annotations, labels, router_type_config = _init_router(router_name)
    route_type = f'{target_type}-subdomain'
    labels = labels or {}
    labels.update(**{
        'ckan-cloud/route-type': route_type,
        'ckan-cloud/route-root-domain': root_domain,
        'ckan-cloud/route-sub-domain': sub_domain,
        'ckan-cloud/route-target-type': target_type,
        'ckan-cloud/route-target-resource-id': target_resource_id,
    })
    spec = {
        'name': route_name,
        'type': route_type,
        'root-domain': root_domain,
        'sub-domain': sub_domain,
        'router_name': router_name,
        'router_type': router_type,
        'route-target-type': target_type,
        'route-target-resource-id': target_resource_id,
    }
    if target_type == 'datapusher':
        labels['ckan-cloud/route-datapusher-name'] = spec['datapusher-name'] = route_spec['datapusher-name']
    elif target_type == 'deis-instance':
        labels['ckan-cloud/route-deis-instance-id'] = spec['deis-instance-id'] = route_spec['deis-instance-id']
    elif target_type == 'backend-url':
        spec['backend-url'] = route_spec['backend-url']
    elif target_type == 'ckan-instance':
        labels['ckan-cloud/route-ckan-instance-id'] = spec['ckan-instance-id'] = route_spec['ckan-instance-id']
    elif target_type == 'app-instance':
        labels['ckan-cloud/route-app-instance-id'] = spec['app-instance-id'] = route_spec['app-instance-id']
    route = kubectl.get_resource('stable.viderum.com/v1', 'CkanCloudRoute', route_name, labels, spec=spec)
    kubectl.apply(route, dry_run=dry_run)


def install_crds():
    """Ensures installaion of the custom resource definitions on the cluster"""
    kubectl.install_crd('ckancloudrouters', 'ckancloudrouter', 'CkanCloudRouter')
    routes_manager.install_crds()


def delete(router_name, router_type=None):
    if router_type:
        router_type_config = ROUTER_TYPES[router_type]
    else:
        router, spec, router_type, annotations, labels, router_type_config = _init_router(router_name)
    router_type_config['manager'].delete(router_name)


def get_datapusher_routes(datapusher_name, edit=False):
    labels = {'ckan-cloud/route-datapusher-name': datapusher_name}
    if edit: kubectl.edit_items_by_labels('CkanCloudRoute', labels)
    else: return kubectl.get_items_by_labels('CkanCloudRoute', labels, required=False)


def get_backend_url_routes(target_resorce_id, edit=False):
    labels = {'ckan-cloud/route-target-resource-id': target_resorce_id}
    if edit: kubectl.edit_items_by_labels('CkanCloudRoute', labels)
    else: return kubectl.get_items_by_labels('CkanCloudRoute', labels, required=False)


def get_deis_instance_routes(deis_instance_id, edit=False):
    labels = {'ckan-cloud/route-deis-instance-id': deis_instance_id}
    if edit: kubectl.edit_items_by_labels('CkanCloudRoute', labels)
    else: return kubectl.get_items_by_labels('CkanCloudRoute', labels, required=False)


def get_ckan_instance_routes(ckan_instance_id, edit=False):
    labels = {'ckan-cloud/route-ckan-instance-id': ckan_instance_id}
    if edit: kubectl.edit_items_by_labels('CkanCloudRoute', labels)
    else: return kubectl.get_items_by_labels('CkanCloudRoute', labels, required=False)


def get_app_instance_routes(app_instance_id, edit=False):
    labels = {'ckan-cloud/route-app-instance-id': app_instance_id}
    if edit: kubectl.edit_items_by_labels('CkanCloudRoute', labels)
    else: return kubectl.get_items_by_labels('CkanCloudRoute', labels, required=False)


def get_all_routes():
    return kubectl.get('CkanCloudRoute')['items']


def get_domain_routes(root_domain=None, sub_domain=None):
    if not root_domain or root_domain == get_default_root_domain():
        root_domain = 'default'
    assert sub_domain or root_domain != 'default', 'cannot delete all routes from default root domain'
    labels = {'ckan-cloud/route-root-domain': root_domain}
    if sub_domain:
        labels['ckan-cloud/route-sub-domain']: sub_domain
    return kubectl.get_items_by_labels('CkanCloudRoute', labels, required=False)


# delete_routes mtehod is dangerous - better to delete manually via kubectl
#
# def delete_routes(routes=None, deis_instance_id=None, backend_url_target_resource_id=None, datapusher_name=None,
#                   root_domain=None, sub_domain=None):
#     if deis_instance_id:
#         assert not routes and not backend_url_target_resource_id and not datapusher_name and not root_domain and not sub_domain
#         routes = get_deis_instance_routes(deis_instance_id)
#     elif backend_url_target_resource_id:
#         assert not routes and not datapusher_name and not root_domain and not sub_domain
#         routes = get_backend_url_routes(backend_url_target_resource_id)
#     elif datapusher_name:
#         assert not routes and not root_domain and not sub_domain
#         routes = get_datapusher_routes(datapusher_name)
#     elif root_domain or sub_domain:
#         assert not routes
#         routes = get_domain_routes(root_domain, sub_domain)
#     assert routes
#     delete_route_names = set()
#     update_router_names = set()
#     for route in routes:
#         delete_route_names.add(route['metadata']['name'])
#         update_router_names.add(route['spec']['router_name'])
#     if len(delete_route_names) > 0:
#         delete_route_names = ' '.join(delete_route_names)
#         kubectl.check_call(f'delete CkanCloudRoute {delete_route_names}')
#         for router_name in update_router_names:
#             update(router_name)


def get_default_infra_router_name():
    return 'infra-1'


def get_route_frontend_hostname(route):
    frontend_hostname = routes_manager.get_frontend_hostname(route)
    if frontend_hostname.endswith('.default'):
        return frontend_hostname.replace('.default', '.' + get_default_root_domain())
    else:
        return frontend_hostname


def get_cloudflare_rate_limits(root_domain):
    from ckan_cloud_operator.providers.routers import manager as routers_manager
    from ckan_cloud_operator import cloudflare
    return cloudflare.get_zone_rate_limits(*routers_manager.get_cloudflare_credentials(), root_domain)


def _get_labels(router_name, router_type):
    return {'ckan-cloud/router-name': router_name, 'ckan-cloud/router-type': router_type}


def _init_router(router_name, router_values=None, required=False):
    router = kubectl.get(f'CkanCloudRouter {router_name}', required=required) if not router_values else router_values
    if router:
        spec = router['spec']
        router_type = spec['type']
        assert router_type in ROUTER_TYPES, f'Unsupported router type: {router_type}'
        router_type_config = ROUTER_TYPES[router_type]
        annotations = CkanRoutersAnnotations(router_name, router)
        labels = _get_labels(router_name, router_type)
        logs.debug_verbose('_init_router', router=router, router_type_config=router_type_config, labels=labels)
        return router, spec, router_type, annotations, labels, router_type_config
    else:
        logs.debug_verbose('_init_router', router=router, router_type_config=None, labels=None)
        return None, None, None, None, None, None


def _get_default_sub_root_domain(sub_domain, root_domain, default_sub_domain_suffix):
    if not root_domain or root_domain == 'default':
        root_domain = 'default'
    if not sub_domain or sub_domain == 'default':
        env_id = get_env_id()
        assert env_id, 'missing env id value'
        sub_domain = f'cc-{env_id}-{default_sub_domain_suffix}'
    _validate_sub_root_domain(sub_domain, root_domain)
    return sub_domain, root_domain


def _validate_sub_root_domain(sub_domain, root_domain):
    env_id = get_env_id()
    assert env_id, 'missing env id value'
    assert env_id == 'p' or root_domain == 'default', 'non-default root domain is not allowed for non production environmnets'
    assert env_id == 'p' or sub_domain.startswith(f'cc-{env_id}-'), 'non-default sub-domains are not allowed for non production environments'



from ckan_cloud_operator import logs
from ckan_cloud_operator import kubectl

from ckan_cloud_operator.config import manager as config_manager


def initialize(log_kwargs=None):
    logs.info('setting label-prefix: ckan-cloud', **(log_kwargs or {}))
    config_manager.set('label-prefix', 'ckan-cloud')


def get_label_prefix(short=False):
    """Returns a global label prefix which should be used to namespace operator objects"""
    return config_manager.get('short-label-prefix' if short else 'label-prefix', required=True)


def get_resource_name(suffix, short=False):
    label_prefix = get_label_prefix(short=short)
    if short:
        return f'{label_prefix}{suffix}'
    else:
        return f'{label_prefix}-{suffix}'


def get_resource_labels(label_suffixes, extra_labels=None):
    label_prefix = get_label_prefix()
    labels = {
        f'{label_prefix}/{label_suffix}': label_value
        for label_suffix, label_value in label_suffixes.items()
    }
    if extra_labels:
        labels.update(**extra_labels)
    return labels


def delete_by_labels(labels, kinds):
    kubectl.delete_items_by_labels(kinds, labels)



import subprocess
from ckan_cloud_operator import kubectl


class DeisCkanInstanceCKAN(object):
    """Manage the CKAN app"""

    def __init__(self, instance):
        self.instance = instance

    def update(self):
        self.instance.annotations.update_status('ckan', 'created', lambda: self._create())

    def run(self, command, *args):
        assert command in ['paster', 'exec', 'logs', 'port-forward']
        command = command.replace('-', '_')
        getattr(self, command)(*args)

    def paster(self, paster_command=None, *paster_args):
        """Run ckan-paster commands on the first CKAN pod using relevant CKAN configuration file"""
        cmd = f'-it -- paster --plugin=ckan'
        if paster_command:
            cmd += f' {paster_command} -c /srv/app/production.ini ' + " ".join(paster_args)
        self.exec(cmd)

    def admin_credentials(self):
        data = kubectl.decode_secret(kubectl.get('secret', 'ckan-env-vars', namespace=self.instance.id))
        return {
            'sysadmin-name': data['CKAN_SYSADMIN_NAME'],
            'sysadmin-password': data['CKAN_SYSADMIN_PASSWORD'],
        }

    def _get_ckan_pod_name(self):
        deployment_data = self.instance.get('deployment')
        latest_pod_name = deployment_data['latest_pod_name']
        if latest_pod_name:
            latest_pods = [pod for pod in deployment_data.get('pods', []) if pod['name'] == latest_pod_name]
            if len(latest_pods) == 1 and len(latest_pods[0].get('errors', [])) == 0:
                return latest_pod_name
        return None

    def exec(self, *args, check_output=False):
        """Execute shell scripts on the first CKAN pod"""
        pod_name = self._get_ckan_pod_name()
        assert pod_name
        return self.instance.kubectl(f'exec {pod_name} ' + " ".join(args), check_output=check_output)

    def logs(self, *args):
        """Run kubectl logs on the first CKAN pod"""
        pod_name = self._get_ckan_pod_name()
        assert pod_name
        self.instance.kubectl(f'logs {pod_name} ' + " ".join(args))

    def port_forward(self, *args):
        """Start port forwarding to the CKAN deployment, using the CKAN varnish port 5000 by default"""
        if len(args) == 0:
            args = ['5000']
        pod_name = self._get_ckan_pod_name()
        self.instance.kubectl(f'port-forward {pod_name} ' + " ".join(args))
        subprocess.check_call(['kubectl', '-n', self.instance.id, 'port-forward', f'deployment/{self.instance.id}', *args])

    def _create(self):
        ckan_init = self.instance.spec.spec.get('ckan', {}).get('init')
        if ckan_init:
            for cmd in ckan_init:
                print('Running ckan init script')
                if cmd[0] == 'paster':
                    print(' '.join(cmd))
                    self.paster(*cmd[1:])
                else:
                    raise ValueError(f'Invalid ckan init cmd: {cmd}')



#### standard provider code ####

# import the correct PROVIDER_SUBMODULE and PROVIDER_ID constants for your provider
from ckan_cloud_operator.providers.db.proxy.pgbouncer.constants import PROVIDER_ID
from ckan_cloud_operator.providers.db.proxy.constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file)
def _config_get_volume_spec(volume_name, is_secret=False, suffix=None): return providers_manager.config_get_volume_spec(PROVIDER_SUBMODULE, PROVIDER_ID, volume_name, is_secret, suffix)

################################
# custom provider code starts here
#

import time
import os
import subprocess
import traceback

from distutils.util import strtobool

from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs

from ckan_cloud_operator.providers.db import manager as db_manager
from ckan_cloud_operator.providers.cluster import manager as cluster_manager


def initialize():
    _apply_config_secret(force=True)
    _apply_service()
    _apply_deployment()
    _set_provider()


def update(wait_updated=False, set_pool_mode=None):
    if set_pool_mode:
        _config_set('pool-mode', set_pool_mode)
    _apply_config_secret()
    if wait_updated:
        logs.info('Waiting 1 minute for pgbouncer secret to be updated...')
        time.sleep(60)
        reload()


def reload():
    deployment_app = _get_resource_labels(for_deployment=True)['app']
    logs.info('Reloading pgbouncers...')
    for pod_name in kubectl.check_output(
            f'get pods -l app={deployment_app} --output=custom-columns=name:.metadata.name --no-headers'
    ).decode().splitlines():
        kubectl.check_call(f'exec {pod_name} -- pgbouncer -q -u pgbouncer -d -R /var/local/pgbouncer/pgbouncer.ini')
        logs.info(f'{pod_name}: PgBouncer Reloaded')


def get_internal_proxy_host_port():
    namespace = cluster_manager.get_operator_namespace_name()
    service_name = _get_resource_name()
    return f'{service_name}.{namespace}', 5432


def get_external_proxy_host_port():
    if strtobool(os.environ.get('CKAN_CLOUD_OPERATOR_USE_PROXY', 'y')):
        host, port = get_external_proxy_forwarded_host_port()
    else:
        host, port = None, None
    return host, port


def get_external_proxy_forwarded_host_port():
    return '127.0.0.1', 5432


def start_port_forward():
    """Starts a local proxy to the cloud SQL instance"""
    print("\nKeep this running in the background\n")
    namespace = cluster_manager.get_operator_namespace_name()
    deployment_name = _get_resource_name()
    subprocess.check_call(f'kubectl -n {namespace} port-forward deployment/{deployment_name} 5432',
                          shell=True)


def _apply_config_secret(force=False):
    update_dbs = {}
    update_users = {}
    try:
        dbs, users = db_manager.get_all_dbs_users()
    except Exception:
        if force:
            traceback.print_exc()
            dbs, users = [], []
        else:
            raise
    for db_name, db_host, db_port in dbs:
        assert db_name not in update_dbs
        update_dbs[db_name] = f'{db_name} = host={db_host} port={db_port} dbname={db_name}'
    for name, password in users:
        assert name not in users
        update_users[name] = password
    pg_bouncer_ini = ["[databases]"]
    for db_name, line in update_dbs.items():
        pg_bouncer_ini.append(line)
    db_admin_user, _, _ = db_manager.get_admin_db_credentials()
    # see https://pgbouncer.github.io/config.html
    pool_mode = _config_get('pool-mode', 'transaction')
    pg_bouncer_ini += [
        "",
        "[pgbouncer]",
        "listen_port = 5432",
        "listen_addr = 0.0.0.0",
        "auth_type = md5",
        "auth_file = /var/local/pgbouncer/users.txt",
        "logfile = /var/log/pgbouncer/pgbouncer.log",
        "pidfile = /var/run/pgbouncer/pgbouncer.pid",
        f"pool_mode = {pool_mode}",
        *([
            "default_pool_size = 8",
            "reserve_pool_size = 8",
            "max_client_conn = 5000",
            "server_round_robin = 1",
            "listen_backlog = 8192",
            "server_idle_timeout = 60",
            "server_lifetime = 600",
        ] if pool_mode == 'transaction' else [
            "default_pool_size = 20",
            "reserve_pool_size = 20",
            "max_client_conn = 5000",
            "server_round_robin = 1",
            "server_fast_close = 1",
            "listen_backlog = 8192",
            "server_lifetime = 5",
        ]),
        f"admin_users = {db_admin_user}",
    ]
    users_txt = []
    for name, password in update_users.items():
        users_txt.append(f'"{name}" "{password}"')
    updated_secret = {
        'pgbouncer.ini': "\n".join(pg_bouncer_ini),
        'users.txt': "\n".join(users_txt)
    }
    _config_set(values=updated_secret, is_secret=True)


def _apply_deployment():
    kubectl.apply(kubectl.get_deployment(
        _get_resource_name(),
        _get_resource_labels(for_deployment=True),
        {
            'replicas': 1,
            'revisionHistoryLimit': 10,
            'strategy': {'type': 'RollingUpdate', },
            'template': {
                'metadata': {
                    'labels': _get_resource_labels(for_deployment=True),
                    'annotations': _get_resource_annotations()
                },
                'spec': {
                    'containers': [
                        {
                            'name': 'pgbouncer',
                            'image': 'viderum/ckan-cloud-operator:pgbouncer',
                            'ports': [{'containerPort': 5432}],
                            'volumeMounts': [
                                {
                                    'name': 'config',
                                    'mountPath': '/var/local/pgbouncer',
                                    'readOnly': True
                                },
                            ],
                            'readinessProbe': {
                                'failureThreshold': 1,
                                'initialDelaySeconds': 5,
                                'periodSeconds': 5,
                                'successThreshold': 1,
                                'tcpSocket': {
                                    'port': 5432
                                },
                                'timeoutSeconds': 5
                            },
                            'resources': {
                                'limits': {
                                    'memory': '2Gi',
                                },
                                'requests': {
                                    'cpu': '0.1',
                                    'memory': '0.2Gi',
                                }
                            }
                        }
                    ],
                    'volumes': [
                        _config_get_volume_spec('config', is_secret=True),
                    ]
                }
            }
        }
    ))


def _apply_service():
    deployment_app = _get_resource_labels(for_deployment=True)['app']
    kubectl.apply(kubectl.get_service(
        _get_resource_name(),
        _get_resource_labels(),
        [5432],
        {'app': deployment_app}
    ))



import json
import subprocess
import re
import functools
import traceback
import requests


from ckan_cloud_operator import logs


class CkanGitlab(object):

    def initialize(self, project, git_branch='master'):
        project_id = self._get_project_id(project)
        print({k: v for k, v in json.loads(self._curl(
            f'projects/{project_id}',
            postjson={'container_registry_enabled': True},
            method='PUT'
        )).items() if k in ['id', 'path_with_namespace', 'container_registry_enabled']})
        for fname in ['.env', 'Dockerfile', '.gitlab-ci.yml']:
            data = self._get_file(project, fname, ref=git_branch, required=False)
            if fname == 'Dockerfile':
                self._update_dockerfile(project_id, data, git_branch)
            if data:
                size_b = len(data)
                assert size_b > 0
                print(f'{project}/{fname}: {size_b} bytes')
            elif fname == '.gitlab-ci.yml':
                print(f'Creating {project}/.gitlab-ci.yml')
                project_id = self._get_project_id(project)
                self._curl(f'projects/{project_id}/repository/files/.gitlab-ci.yml', postjson={
                    'branch': git_branch, 'author_email': 'admin@ckan-cloud-operator',
                    'author_name': 'ckan-cloud-operator',
                    'content': self._get_gitlab_ci_yml(git_branch=git_branch), 'commit_message': 'Add .gitlab-ci.yml'
                })
            else:
                print(f'Repository is missing {project}/{fname}')

    def is_ready(self, project):
        project_id = self._get_project_id(project)
        project_pipelines = json.loads(self._curl(
            f'projects/{project_id}/pipelines'
        ))
        print(project_pipelines)
        success =  [p for p in project_pipelines if p['status'] == 'success']
        pending =  [p for p in project_pipelines if p['status'] == 'pending']
        failed =  [p for p in project_pipelines if p['status'] == 'failed']
        running =  [p for p in project_pipelines if p['status'] == 'running']
        others = [p for p in project_pipelines if p['status'] not in ['failed', 'success', 'pending', 'running']]
        assert len(others) == 0, f'unknown status: {others}'
        if len(success) > 0 and len(pending) == 0:
            print(f'successful pipelines: {success}')
            print(f'pending pipelines: {pending}')
            print(f'failed pipelines: {failed}')
            print(f'running pipelines: {running}')
            return True
        else:
            return False

    def get_envvars(self, project, git_branch='master'):
        return self._parse_dotenv(self._get_file(project, '.env', ref=git_branch))

    def _get_updated_dockerfile(self, data):
        needs_update = False
        got_upgraded = False
        lines = []
        for line in data.splitlines():
            if 'pip install --upgrade pip' in line:
                got_upgraded = True
            elif not got_upgraded and line.startswith('RUN pip install '):
                logs.warning('please add pip install --upgrade pip manually to the Dockerfile (if needed)')
                # needs_update = True
                # line = 'RUN pip install --upgrade pip==18.1 && pip install {}'.format(line[15:])
            lines.append(line)
        from ckan_cloud_operator.providers.ckan import manager as ckan_manager
        search_replace_needs_update, lines = ckan_manager.gitlab_search_replace(lines)
        if needs_update or search_replace_needs_update:
            return needs_update, search_replace_needs_update, '{}\n'.format('\n'.join(lines))
        else:
            return False, False, None

    def _update_dockerfile(self, project_id, data, git_branch='master'):
        pip_update, gitlab_search_replace_update, updated_data = self._get_updated_dockerfile(data)
        if updated_data:
            commit_messages = []
            if pip_update: commit_messages.append('Add pip install --upgrade pip to Dockerfile')
            if gitlab_search_replace_update: commit_messages.append('replace values for migration')
            self._curl(f'projects/{project_id}/repository/files/Dockerfile', postjson={
                'branch': git_branch, 'author_email': 'admin@ckan-cloud-operator',
                'author_name': 'ckan-cloud-operator',
                'content': updated_data, 'commit_message': ', '.join(commit_messages)
            }, method='PUT')

    def _get_file(self, project, file, ref='master', required=True):
        try:
            project_id = self._get_project_id(project)
            file = file.replace('/', '%2F')
            return self._curl(f'projects/{project_id}/repository/files/{file}/raw?ref={ref}')
        except Exception:
            if required:
                raise
            else:
                traceback.print_exc()
                return None

    def _curl(self, urlpart, postjson=None, method='POST', token_name=None, download_filename=None):
        from ckan_cloud_operator.providers.ckan import manager as ckan_manager
        gitlab_token = ckan_manager.gitlab_token(token_name)
        if postjson:
            r = requests.request(
                method,
                f'https://gitlab.com/api/v4/{urlpart}',
                headers={
                    'PRIVATE-TOKEN': gitlab_token
                },
                json=postjson
            )
            assert r.status_code in [200, 201], r.text
            return r.text
        elif download_filename:
            cmd = f'curl -f -s --header "PRIVATE-TOKEN: {gitlab_token}" https://gitlab.com/api/v4/{urlpart} > {download_filename}'
            subprocess.check_call(cmd, shell=True)
        else:
            cmd = ['curl', '-f', '-s',
                   '--header', f'PRIVATE-TOKEN: {gitlab_token}',
                   f'https://gitlab.com/api/v4/{urlpart}']
            return subprocess.check_output(cmd).decode()

    @functools.lru_cache()
    def _get_project_id(self, project):
        gitlab_project_encoded = project.replace('/', '%2F')
        return json.loads(self._curl(f'projects/{gitlab_project_encoded}'))['id']

    def _parse_dotenv(self, content, overrides=None):
        # https://github.com/joke2k/django-environ/blob/develop/environ/environ.py
        env = {}
        for line in content.splitlines():
            m1 = re.match(r'\A(?:export )?([A-Za-z_0-9]+)=(.*)\Z', line)
            if m1:
                key, val = m1.group(1), m1.group(2)
                m2 = re.match(r"\A'(.*)'\Z", val)
                if m2:
                    val = m2.group(1)
                m3 = re.match(r'\A"(.*)"\Z', val)
                if m3:
                    val = re.sub(r'\\(.)', r'\1', m3.group(1))
                env.setdefault(key, str(val))
        if overrides:
            for key, value in overrides.items():
                env.setdefault(key, value)
        return env

    def _get_gitlab_ci_yml(self, git_branch='master'):
        return """image: docker:latest

services:
  - docker:dind

before_script:
  - docker login -u "$CI_REGISTRY_USER" -p "$CI_REGISTRY_PASSWORD" $CI_REGISTRY

build-{git_branch}:
  stage: build
  script:
    - docker build --pull -t "$CI_REGISTRY_IMAGE" .
    - docker push "$CI_REGISTRY_IMAGE"
  only:
    - {git_branch}

build:
  stage: build
  script:
    - docker build --pull -t "$CI_REGISTRY_IMAGE:$CI_COMMIT_REF_SLUG" .
    - docker push "$CI_REGISTRY_IMAGE:$CI_COMMIT_REF_SLUG"
  except:
    - {git_branch}""".format(git_branch=git_branch)



import click
import yaml

from ckan_cloud_operator import logs
from . import manager


@click.group()
def users():
    pass


@users.command()
@click.option('--interactive', is_flag=True)
def initialize(interactive):
    manager.initialize(interactive=interactive)
    logs.exit_great_success()


@users.command()
@click.argument('USER_NAME')
def get_kubeconfig(user_name):
    """Get kubeconfig file for a user"""
    print(yaml.dump(manager.get_kubeconfig(user_name), default_flow_style=False))
    logs.exit_great_success(quiet=True)


@users.command()
@click.argument('USER_NAME')
@click.option('--role')
def create(user_name, role):
    """Create a user with the given role"""
    manager.create(user_name, role)
    manager.update(user_name)
    logs.exit_great_success()



import click
import yaml

from ckan_cloud_operator import logs

from . import manager


@click.group()
def storage():
    """Manage CKAN Storage"""
    pass


@storage.command()
@click.option('--interactive', is_flag=True)
def initialize(interactive):
    manager.initialize(interactive=interactive)
    logs.exit_great_success()


@storage.command()
def deis_minio_bucket_policy():
    print(manager.get_deis_minio_bucket_policy())
    logs.exit_great_success(quiet=True)



from ckan_cloud_operator import datapushers


from ckan_cloud_operator.routers.routes.backend_url_subdomain import (
    get_frontend_hostname,
    get_domain_parts,
    get_default_root_domain,
    get_route,
    _init_route
)


def get_backend_url(route):
    name, spec = _init_route(route)
    datapusher_name = spec['datapusher-name']
    return datapushers.get_service_url(datapusher_name)


def pre_deployment_hook(route, labels):
    name, spec = _init_route(route)
    datapusher_name = spec['datapusher-name']
    print(f'updating route name {name} for datapusher name {datapusher_name}')
    datapushers.update_service(datapusher_name, labels)



#### standard provider code ####

from .constants import PROVIDER_ID
from ..constants import PROVIDER_SUBMODULE
from ckan_cloud_operator.providers.cluster import manager as cluster_manager

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False, suffix=None): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment, suffix=suffix)
def _get_resource_annotations(suffix=None, with_timestamp=False): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix, with_timestamp=with_timestamp)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False, interactive=True): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file, interactive)

################################
# custom provider code starts here
#

import subprocess
import yaml
import json
import time
import os

from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs
from ckan_cloud_operator.config import manager as config_manager
from ckan_cloud_operator.providers.cluster import manager as cluster_manager

from .constants import LOG4J_PROPERTIES, SOLR_CONFIG_XML


def start_zoonavigator_port_forward():
    connection_string = ','.join(yaml.load(_config_get('zk-host-names')))
    print("\nStarting port forward to zoonavigator\n"
          "\nUse the following connection string:\n"
          f"\n  {connection_string}\n"
          "\nhttp://localhost:8000/\n")
    namespace = cluster_manager.get_operator_namespace_name()
    deployment_name = _get_resource_name('zoonavigator')
    subprocess.check_call(f'kubectl -n {namespace} port-forward deployment/{deployment_name} 8000', shell=True)


def start_solrcloud_port_forward(suffix='sc-0'):
    namespace = cluster_manager.get_operator_namespace_name()
    deployment_name = _get_resource_name(suffix)
    subprocess.check_call(f'kubectl -n {namespace} port-forward deployment/{deployment_name} 8983', shell=True)


def get_internal_http_endpoint():
    solrcloud_host_name = _config_get('sc-main-host-name', required=True)
    namespace = cluster_manager.get_operator_namespace_name()
    return f'http://{solrcloud_host_name}.{namespace}.svc.cluster.local:8983/solr'


def solr_curl(path, required=False, debug=False, max_retries=15):
    deployment_name = _get_resource_name(_get_sc_suffixes()[0])
    if debug:
        kubectl.check_call(f'exec deployment-pod::{deployment_name} -- curl \'localhost:8983/solr{path}\'',
                           use_first_pod=True)
    else:
        exitcode, output = kubectl.getstatusoutput(f'exec deployment-pod::{deployment_name} -- curl -s -f \'localhost:8983/solr{path}\'',
                           use_first_pod=True)
        if exitcode == 0:
            return output
        elif required:
            if max_retries > 0:
                logs.info(f'Failed to run solr curl: localhost:8983/solr{path} - retring in 30 seconds')
                time.sleep(30)
                solr_curl(path, required=required, debug=debug, max_retries=max_retries-1)
            logs.critical(output)
            raise Exception(f'Failed to run solr curl: localhost:8983/solr{path}')
        else:
            logs.warning(output)
            return False


def initialize(interactive=False, dry_run=False):
    if cluster_manager.get_provider_id() == 'minikube':
        config_manager.set('container-spec-overrides', '{"resources":{"limits":{"memory":"1Gi"}}}',
            configmap_name='ckan-cloud-provider-solr-solrcloud-sc-config')

    solr_resources = config_manager.interactive_set(
        {
            'sc-cpu': '1',
            'sc-mem': '1Gi',
            'zk-cpu': '0.2',
            'zk-mem': '200Mi',
            'zn-cpu': '0.01',
            'zn-mem': '0.01Gi',
            'sc-cpu-limit': '2.5',
            'sc-mem-limit': '8Gi',
            'zk-cpu-limit': '0.5',
            'zk-mem-limit': '200Mi',
            'zn-cpu-limit': '0.5',
            'zn-mem-limit': '0.5Gi',
        },
        secret_name='solr-config',
        interactive=interactive
    )

    zk_host_names = initialize_zookeeper(interactive, dry_run=dry_run)

    _config_set('zk-host-names', yaml.dump(zk_host_names, default_flow_style=False))
    logs.info(f'Initialized zookeeper: {zk_host_names}')

    zoonavigator_deployment_name = _apply_zoonavigator_deployment(dry_run=dry_run)
    logs.info(f'Initialized zoonavigator: {zoonavigator_deployment_name}')

    sc_host_names = initialize_solrcloud(zk_host_names, pause_deployment=False, interactive=interactive, dry_run=dry_run)
    _config_set('sc-host-names', yaml.dump(sc_host_names, default_flow_style=False))
    logs.info(f'Initialized solrcloud: {sc_host_names}')

    solrcloud_host_name = _apply_solrcloud_service(dry_run=dry_run)
    _config_set('sc-main-host-name', solrcloud_host_name)
    logs.info(f'Initialized solrcloud service: {solrcloud_host_name}')

    expected_running = len(sc_host_names) + len(zk_host_names) + 1
    RETRIES = 40 # ~20 minutes
    for retry in range(RETRIES):
        pods = kubectl.get('pods')
        running = len([x for x in pods['items']
                       if x['status']['phase'] == 'Running' and x['metadata']['labels']['app'].startswith(_get_resource_labels(for_deployment=True)['app'])])
        time.sleep(45)
        logs.info('Waiting for SolrCloud to start... %d/%d' % (running, expected_running))
        for x in pods['items']:
            logs.info('  - %-10s | %s: %s' % (x['metadata'].get('labels', {}).get('app'), x['metadata']['name'], x['status']['phase']))

        if running == expected_running:
            break
        assert retry < RETRIES - 1, 'Gave up on waiting for SolrCloud'

    _set_provider()


def initialize_zookeeper(interactive=False, dry_run=False):
    headless_service_name = _apply_zookeeper_headless_service(dry_run=dry_run)
    zk_instances = {suffix: {
        'host_name': suffix,
        'volume_spec': _get_or_create_volume(suffix, disk_size_gb=20, dry_run=dry_run, zone=zone),
    } for zone, suffix in enumerate(_get_zk_suffixes())}
    zk_host_names = [zk['host_name'] for zk in zk_instances.values()]
    zk_configmap_name = _apply_zookeeper_configmap(zk_host_names)
    for zk_suffix, zk in zk_instances.items():
        _apply_zookeeper_deployment(zk_suffix, zk['volume_spec'], zk_configmap_name, headless_service_name, dry_run=dry_run)
    namespace = cluster_manager.get_operator_namespace_name()
    return [f'{h}.{headless_service_name}.{namespace}.svc.cluster.local:2181' for h in zk_host_names]


def initialize_solrcloud(zk_host_names, pause_deployment, interactive=False, dry_run=False):
    sc_logs_configmap_name = _apply_solrcloud_logs_configmap()
    headless_service_name = _apply_solrcloud_headless_service(dry_run=dry_run)
    sc_instances = {suffix: {
        'host_name': suffix,
        'volume_spec': _get_or_create_volume(suffix, disk_size_gb=100, dry_run=dry_run, zone=zone)
    } for zone, suffix in enumerate(_get_sc_suffixes())}
    sc_host_names = [sc['host_name'] for sc in sc_instances.values()]
    sc_configmap_name = _apply_solrcloud_configmap(zk_host_names)
    for sc_suffix, sc in sc_instances.items():
        _apply_solrcloud_deployment(sc_suffix, sc['volume_spec'], sc_configmap_name, sc_logs_configmap_name, headless_service_name,
                                    pause_deployment, dry_run=dry_run)
    return sc_host_names


def _get_zk_suffixes():
    if cluster_manager.get_provider_id() != 'minikube':
        return ['zk-0', 'zk-1', 'zk-2']
    else:
        return ['zk-0']


def _get_sc_suffixes():
    if cluster_manager.get_provider_id() != 'minikube':
        return ['sc-3', 'sc-4', 'sc-5']
    else:
        return ['sc-3']


def _apply_zookeeper_configmap(zk_host_names):
    zk_configmap_suffix = 'zk-config'
    _config_set(
        values={
            'ZK_REPLICAS': str(len(zk_host_names)),
            'ZK_ENSEMBLE': ';'.join(zk_host_names),
            'ZK_HEAP_SIZE': '2G',
            'ZK_TICK_TIME': '2000',
            'ZK_INIT_LIMIT': '10',
            'ZK_SYNC_LIMIT': '2000',
            'ZK_MAX_CLIENT_CNXNS': '60',
            'ZK_SNAP_RETAIN_COUNT': '3',
            'ZK_PURGE_INTERVAL': '1',
            'ZK_CLIENT_PORT': '2181',
            'ZK_SERVER_PORT': '2888',
            'ZK_ELECTION_PORT': '3888',
            'ZK_LOG_LEVEL': 'WARN',
        },
        suffix=zk_configmap_suffix
    )
    return _get_resource_name(suffix=zk_configmap_suffix)


def _apply_solrcloud_configmap(zk_host_names):
    sc_configmap_suffix = 'sc-config'
    _config_set(
        values={
            'ZK_HOST': ','.join(zk_host_names),
            'SOLR_HOME': '/data/solr',
            'SOLR_OPTS': '-Dlog4j.configuration=file:///logconfig/log4j.properties -Ddisable.configEdit=true',
        },
        suffix=sc_configmap_suffix
    )
    return _get_resource_name(suffix=sc_configmap_suffix)


def _apply_solrcloud_logs_configmap():
    configmap_suffix = 'sc-logs-config'
    _config_set(
        values={
            'log4j.properties': LOG4J_PROPERTIES,
        },
        suffix=configmap_suffix
    )
    return _get_resource_name(suffix=configmap_suffix)


def _get_volume_pod_scheduling(volume_spec, app_in):
    node_selector = volume_spec.pop('nodeSelector', None)
    return {
        **({'nodeSelector': node_selector} if node_selector else {}),
        'affinity': {
            'podAntiAffinity': {
                'requiredDuringSchedulingIgnoredDuringExecution': [
                    {
                        'labelSelector': {'matchExpressions': [
                            {'key': 'app', 'operator': 'In', 'values': [app_in]}
                        ]},
                        'topologyKey': 'kubernetes.io/hostname'
                    }
                ]
            }
        }
    }


def _apply_zookeeper_deployment(suffix, volume_spec, zookeeper_configmap_name, headless_service_name, dry_run=False):
    cpu_req = config_manager.get('zk-cpu', secret_name='solr-config')
    mem_req = config_manager.get('zk-mem', secret_name='solr-config')
    cpu_lim = config_manager.get('zk-cpu-limit', secret_name='solr-config')
    mem_lim = config_manager.get('zk-mem-limit', secret_name='solr-config')
    kubectl.apply(kubectl.get_deployment(
        _get_resource_name(suffix),
        _get_resource_labels(for_deployment=True, suffix='zk'),
        {
            'replicas': 1,
            'revisionHistoryLimit': 2,
            'strategy': {'type': 'Recreate', },
            'selector': {
                'matchLabels': _get_resource_labels(for_deployment=True, suffix='zk'),
            },
            'template': {
                'metadata': {
                    'labels': _get_resource_labels(for_deployment=True, suffix='zk'),
                    'annotations': _get_resource_annotations()
                },
                'spec': {
                    'hostname': suffix,
                    'subdomain': headless_service_name,
                    **_get_volume_pod_scheduling(
                        volume_spec,
                        _get_resource_labels(for_deployment=True, suffix='zk')['app']
                    ),
                    'containers': [
                        {
                            'name': 'zk',
                            'command': ['sh', '-c', 'zkGenConfig.sh && zkServer.sh start-foreground'],
                            'envFrom': [{'configMapRef': {'name': zookeeper_configmap_name}}],
                            'env': [
                                {'name': 'SOLR_HOST', 'valueFrom': {'fieldRef': {'apiVersion': 'v1', 'fieldPath': 'status.podIP'}}}
                            ],
                            'image': 'gcr.io/google_samples/k8szk:v3',
                            'livenessProbe': {
                                'exec': {'command': ['zkOk.sh']},
                                'failureThreshold': 3, 'initialDelaySeconds': 15, 'periodSeconds': 10,
                                'successThreshold': 1, 'timeoutSeconds': 5
                            },
                            'ports': [
                                {'containerPort': 2181, 'name': 'client', 'protocol': 'TCP'},
                                {'containerPort': 2888, 'name': 'server', 'protocol': 'TCP'},
                                {'containerPort': 3888, 'name': 'leader-election', 'protocol': 'TCP'}
                            ],
                            'readinessProbe': {
                                'exec': {'command': ['zkOk.sh']},
                                'failureThreshold': 3, 'initialDelaySeconds': 15, 'periodSeconds': 10,
                                'successThreshold': 1, 'timeoutSeconds': 5
                            },
                            'resources': {'requests': {'cpu': cpu_req, 'memory': mem_req}, 'limits': {'cpu': cpu_lim, 'memory': mem_lim}},
                            'volumeMounts': [
                                {'mountPath': '/var/lib/zookeeper', 'name': 'datadir'},
                            ],
                        }
                    ],
                    'volumes': [
                        dict(volume_spec, name='datadir')
                    ]
                }
            }
        },
        with_timestamp=False
    ), dry_run=dry_run)


def _apply_zoonavigator_deployment(dry_run=False):
    cpu_req = config_manager.get('zn-cpu', secret_name='solr-config')
    mem_req = config_manager.get('zn-mem', secret_name='solr-config')
    cpu_lim = config_manager.get('zn-cpu-limit', secret_name='solr-config')
    mem_lim = config_manager.get('zn-mem-limit', secret_name='solr-config')

    suffix = 'zoonavigator'
    deployment_name = _get_resource_name(suffix)
    kubectl.apply(kubectl.get_deployment(
        deployment_name,
        _get_resource_labels(for_deployment=True, suffix=suffix),
        {
            'replicas': 1,
            'revisionHistoryLimit': 2,
            'selector': {
                'matchLabels': _get_resource_labels(for_deployment=True, suffix=suffix),
            },
            'template': {
                'metadata': {
                    'labels': _get_resource_labels(for_deployment=True, suffix=suffix),
                    'annotations': _get_resource_annotations()
                },
                'spec': {
                    'containers': [
                        {
                            'env': [
                                {'name': 'API_HOST', 'value': 'localhost'},
                                {'name': 'API_PORT', 'value': '9000'},
                                {'name': 'WEB_HTTP_PORT', 'value': '8000'}
                            ],
                            'image': 'elkozmon/zoonavigator-web:0.5.0',
                            'name': 'zoonavigator-web',
                            'ports': [
                                {'containerPort': 8000, 'name': '8000tcp02', 'protocol': 'TCP'}
                            ],
                            'resources': {}
                        }, {
                            'env': [
                                {'name': 'API_HTTP_PORT', 'value': '9000'}
                            ],
                            'image': 'elkozmon/zoonavigator-api:0.5.0',
                            'name': 'zoonavigator-api',
                            'resources': {'requests': {'cpu': cpu_req, 'memory': mem_req}, 'limits': {'memory': mem_lim, 'cpu': cpu_lim}},
                        }
                    ]
                }
            }
        }
    ), dry_run=dry_run)
    return deployment_name


def _apply_solrcloud_deployment(suffix, volume_spec, configmap_name, log_configmap_name, headless_service_name, pause_deployment, dry_run=False):
    cpu_req = config_manager.get('sc-cpu', secret_name='solr-config')
    mem_req = config_manager.get('sc-mem', secret_name='solr-config')
    cpu_lim = config_manager.get('sc-cpu-limit', secret_name='solr-config')
    mem_lim = config_manager.get('sc-mem-limit', secret_name='solr-config')

    namespace = cluster_manager.get_operator_namespace_name()
    container_spec_overrides = config_manager.get('container-spec-overrides', configmap_name='ckan-cloud-provider-solr-solrcloud-sc-config',
                                                  required=False, default=None)
    resources = {'requests': {'cpu': cpu_req, 'memory': mem_req}, 'limits': {'cpu': cpu_lim, 'memory': mem_lim}} if not container_spec_overrides else {}
    kubectl.apply(kubectl.get_deployment(
        _get_resource_name(suffix),
        _get_resource_labels(for_deployment=True, suffix='sc'),
        {
            'replicas': 1,
            'revisionHistoryLimit': 2,
            'strategy': {'type': 'Recreate', },
            'selector': {
                'matchLabels': _get_resource_labels(for_deployment=True, suffix='sc'),
            },
            'template': {
                'metadata': {
                    'labels': _get_resource_labels(for_deployment=True, suffix='sc'),
                    'annotations': _get_resource_annotations()
                },
                'spec': {
                    'hostname': suffix,
                    'subdomain': headless_service_name,
                    **_get_volume_pod_scheduling(
                        volume_spec,
                        _get_resource_labels(for_deployment=True, suffix='sc')['app']
                    ),
                    'initContainers': [
                        {
                            'name': 'init',
                            'image': 'alpine',
                            'command': [
                                "sh", "-c",
                                f"""
                                    if [ -e /data/solr/solr.xml ]; then
                                        echo /data/solr/solr.xml already exists, will not recreate
                                    else
                                        echo creating /data/solr/solr.xml &&\
                                        mkdir -p /data/solr &&\
                                        echo \'{SOLR_CONFIG_XML}\' > /data/solr/solr.xml
                                    fi &&\
                                    echo Setting permissions to solr user/group 8983:8983 on /data/solr &&\
                                    chown -R 8983:8983 /data/solr &&\
                                    echo init completed successfully
                                """
                            ],
                            'securityContext': {
                                'runAsUser': 0
                            },
                            'volumeMounts': [
                                {'mountPath': '/data', 'name': 'datadir'},
                            ]
                        }
                    ],
                    'containers': [
                        {
                            'name': 'sc',
                            'envFrom': [{'configMapRef': {'name': configmap_name}}],
                            'env': [
                                {'name': 'SOLR_HOST', 'value': f'{suffix}.{headless_service_name}.{namespace}.svc.cluster.local'}
                            ],
                            **({
                                'command': ['sh', '-c', 'sleep 86400']
                            } if pause_deployment else {
                                'livenessProbe': {
                                    'exec': {'command': ['/opt/solr/bin/solr', 'status']},
                                    'failureThreshold': 3, 'initialDelaySeconds': 15, 'periodSeconds': 10,
                                    'successThreshold': 1, 'timeoutSeconds': 5
                                },
                                'readinessProbe': {
                                    'exec': {'command': ['/opt/solr/bin/solr', 'status']},
                                    'failureThreshold': 3, 'initialDelaySeconds': 15, 'periodSeconds': 10,
                                    'successThreshold': 1, 'timeoutSeconds': 5
                                },
                            }),
                            'image': 'solr:5.5.5',
                            'ports': [
                                {'containerPort': 8983, 'name': 'solr', 'protocol': 'TCP'},
                                {'containerPort': 7983, 'name': 'stop', 'protocol': 'TCP'},
                                {'containerPort': 18983, 'name': 'rmi', 'protocol': 'TCP'}
                            ],
                            'volumeMounts': [
                                {'mountPath': '/data', 'name': 'datadir'},
                                {'mountPath': '/logconfig', 'name': 'logconfig'}
                            ],
                            **({'resources': resources} if resources else {}),
                            **(json.loads(container_spec_overrides) if container_spec_overrides else {})
                        }
                    ],
                    'volumes': [
                        {'configMap': {'defaultMode': 420, 'name': log_configmap_name}, 'name': 'logconfig'},
                        dict(volume_spec, name='datadir')
                    ]
                }
            }
        },
        with_timestamp=False
    ), dry_run=dry_run)


def _apply_zookeeper_headless_service(dry_run=False):
    headless_service_name = _get_resource_name('zk-headless')
    kubectl.apply(kubectl.get_resource(
        'v1', 'Service',
        headless_service_name,
        _get_resource_labels(suffix='zk-headless'),
        spec={
            'clusterIP': 'None',
            'ports': [
                {'name': 'client', 'port': 2181, 'protocol': 'TCP', 'targetPort': 2181},
                {'name': 'server', 'port': 2888, 'protocol': 'TCP', 'targetPort': 2888},
                {'name': 'leader-election', 'port': 3888, 'protocol': 'TCP', 'targetPort': 3888}
            ],
            'selector': {
                'app': _get_resource_labels(for_deployment=True, suffix='zk')['app']
            }
        }
    ), dry_run=dry_run)
    return headless_service_name


def _apply_solrcloud_headless_service(dry_run=False):
    headless_service_name = _get_resource_name('sc-headless')
    kubectl.apply(kubectl.get_resource(
        'v1', 'Service',
        headless_service_name,
        _get_resource_labels(suffix='sc-headless'),
        spec={
            'clusterIP': 'None',
            'ports': [
                {'name': 'solr', 'port': 8983, 'protocol': 'TCP', 'targetPort': 8983},
                {'name': 'stop', 'port': 7983, 'protocol': 'TCP', 'targetPort': 7983},
                {'name': 'rmi', 'port': 18983, 'protocol': 'TCP', 'targetPort': 18983}
            ],
            'selector': {
                'app': _get_resource_labels(for_deployment=True, suffix='sc')['app']
            }
        }
    ), dry_run=dry_run)
    return headless_service_name


def _apply_solrcloud_service(dry_run=False):
    service_name = _get_resource_name('sc')
    kubectl.apply(kubectl.get_resource(
        'v1', 'Service',
        service_name,
        _get_resource_labels(suffix='sc'),
        spec={
            'ports': [
                {'name': 'solr', 'port': 8983, 'protocol': 'TCP', 'targetPort': 8983},
            ],
            'selector': {
                'app': _get_resource_labels(for_deployment=True, suffix='sc')['app']
            }
        }
    ), dry_run=dry_run)
    return service_name


def _get_or_create_volume(suffix, disk_size_gb, dry_run=False, zone=0):
    volume_spec_config_key = f'volume-spec-{suffix}'
    volume_spec = _config_get(volume_spec_config_key, required=False)
    if volume_spec:
        volume_spec = yaml.load(volume_spec)
    else:
        assert not dry_run, 'creating a new volume is not supported for dry_run'
        from ckan_cloud_operator.providers.cluster import manager as cluster_manager
        volume_spec = cluster_manager.create_volume(disk_size_gb, _get_resource_labels(suffix=suffix), zone=zone)
        _config_set(volume_spec_config_key, yaml.dump(volume_spec, default_flow_style=False))
    if dry_run:
        print(yaml.dump(volume_spec, default_flow_style=False))
    return volume_spec



from ckan_cloud_operator.providers import manager as providers_manager
from .constants import PROVIDER_SUBMODULE
from .gcloudsql.constants import PROVIDER_ID as gcloudsql_provider_id


def initialize():
    get_provider(default=gcloudsql_provider_id).initialize()


def start_port_forward(db_prefix=None):
    get_provider().start_port_forward(db_prefix=db_prefix)


def update(wait_updated=False, set_pool_mode=None):
    get_provider().update(wait_updated=wait_updated, set_pool_mode=set_pool_mode)


def reload():
    get_provider().reload()


def get_provider(default=None, required=True):
    return providers_manager.get_provider(PROVIDER_SUBMODULE, default=default, required=required)



import subprocess
import yaml
from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs
from ckan_cloud_operator.gitlab import CkanGitlab
from ckan_cloud_operator.deis_ckan import datapusher
from ckan_cloud_operator.providers.db import manager as db_manager


class DeisCkanInstanceEnvvars(object):

    def __init__(self, instance):
        self.instance = instance
        self.solr_spec = self.instance.spec.solrCloudCollection
        self.site_url = None

    def _apply_instance_envvars_overrides(self, envvars):
        if 'overrides' in self.instance.spec.envvars:
            print('Applying instance envvars overrides')
            envvars.update(**self.instance.spec.envvars['overrides'])

    def _update(self):
        spec = self.instance.spec
        db_name = spec.db['name']
        db_password = self.instance.annotations.get_secret('databasePassword')
        datastore_name = spec.datastore['name']
        datastore_password = self.instance.annotations.get_secret('datastorePassword')
        datastore_ro_user = self.instance.annotations.get_secret('datastoreReadonlyUser')
        datastore_ro_password = self.instance.annotations.get_secret('datatastoreReadonlyPassword')
        no_db_proxy = True
        # db_no_db_proxy = spec.db.get('no-db-proxy') == 'yes'
        # datastore_no_db_proxy = spec.datastore.get('no-db-proxy') == 'yes'
        # if db_no_db_proxy or datastore_no_db_proxy:
        #     assert db_no_db_proxy and datastore_no_db_proxy, 'must set both DB and datastore with no-db-proxy'
        #     no_db_proxy = True
        # else:
        #     no_db_proxy = False
        from ckan_cloud_operator.providers.solr import manager as solr_manager
        solr_http_endpoint = solr_manager.get_internal_http_endpoint()
        solr_collection_name = spec.solrCloudCollection['name']
        if 'fromSecret' in spec.envvars:
            envvars = kubectl.get(f'secret {spec.envvars["fromSecret"]}')
            envvars = yaml.load(kubectl.decode_secret(envvars, 'envvars.yaml'))
        elif 'fromGitlab' in spec.envvars:
            envvars = CkanGitlab().get_envvars(spec.envvars['fromGitlab'])
        else:
            raise Exception(f'invalid envvars spec: {spec.envvars}')
        from ckan_cloud_operator.providers.storage import manager as storage_manager
        storage_hostname, storage_access_key, storage_secret_key = storage_manager.get_provider().get_credentials()
        storage_path_parts = spec.storage['path'].strip('/').split('/')
        storage_bucket = storage_path_parts[0]
        storage_path = '/'.join(storage_path_parts[1:])
        if no_db_proxy:
            postgres_host, postgres_port = db_manager.get_internal_unproxied_db_host_port(db_prefix=spec.db.get('dbPrefix') or '')
            logs.info(f'Bypassing db proxy, connecting to DB directly: {postgres_host}:{postgres_port}')
        else:
            postgres_host, postgres_port = db_manager.get_internal_proxy_host_port()
            logs.info(f'Connecting to DB proxy: {postgres_host}:{postgres_port}')
        envvars.update(
            CKAN_SQLALCHEMY_URL=f"postgresql://{db_name}:{db_password}@{postgres_host}:{postgres_port}/{db_name}",
            CKAN___BEAKER__SESSION__URL=f"postgresql://{db_name}:{db_password}@{postgres_host}:{postgres_port}/{db_name}",
            CKAN__DATASTORE__READ_URL=f"postgresql://{datastore_ro_user}:{datastore_ro_password}@{postgres_host}:{postgres_port}/{datastore_name}",
            CKAN__DATASTORE__WRITE_URL=f"postgresql://{datastore_name}:{datastore_password}@{postgres_host}:{postgres_port}/{datastore_name}",
            CKAN_SOLR_URL=f"{solr_http_endpoint}/{solr_collection_name}",
            CKANEXT__S3FILESTORE__AWS_STORAGE_PATH=storage_path,
            CKANEXT__S3FILESTORE__AWS_ACCESS_KEY_ID=storage_access_key,
            CKANEXT__S3FILESTORE__AWS_SECRET_ACCESS_KEY=storage_secret_key,
            CKANEXT__S3FILESTORE__AWS_BUCKET_NAME=storage_bucket,
            CKANEXT__S3FILESTORE__HOST_NAME=f'https://{storage_hostname}/',
            CKANEXT__S3FILESTORE__REGION_NAME='us-east-1',
            CKANEXT__S3FILESTORE__SIGNATURE_VERSION='s3v4',
            CKAN__DATAPUSHER__URL=datapusher.get_datapusher_url(envvars.get('CKAN__DATAPUSHER__URL')),
        )
        from ckan_cloud_operator.providers.ckan import manager as ckan_manager
        ckan_manager.update_deis_instance_envvars(self.instance, envvars)
        assert envvars['CKAN_SITE_ID'] and envvars['CKAN_SITE_URL'] and envvars['CKAN_SQLALCHEMY_URL']
        # print(yaml.dump(envvars, default_flow_style=False))
        self._apply_instance_envvars_overrides(envvars)
        envvars = {
            k: ('' if v is None else v)
            for k,v
            in envvars.items()
        }
        kubectl.update_secret('ckan-envvars', envvars, namespace=self.instance.id)
        self.site_url = envvars.get('CKAN_SITE_URL')

    def update(self):
        self.instance.annotations.update_status('envvars', 'created', lambda: self._update(), force_update=True)

    def delete(self):
        print('Deleting instance envvars secret')
        subprocess.check_call(f'kubectl -n {self.instance.id} delete secret/ckan-envvars', shell=True)

    def get(self, full=False):
        exitcode, output = subprocess.getstatusoutput(f'kubectl -n {self.instance.id} get secret/ckan-envvars -o yaml')
        if exitcode == 0:
            secret = kubectl.decode_secret(yaml.load(output))
            res = {'ready': 'CKAN_SITE_URL' in secret}
            if full:
                res['envvars'] = secret
        else:
            res = {'ready': False, 'error': output}
        return res



import click
import yaml

from ckan_cloud_operator import logs

from . import manager


@click.group()
def storage():
    """Manage the centralized storage"""
    pass


@storage.command()
@click.option('--interactive', is_flag=True)
@click.option('--provider-id')
@click.option('--storage-suffix')
@click.option('--disk-name')
@click.option('--dry-run', is_flag=True)
def initialize(interactive, provider_id, storage_suffix, disk_name, dry_run):
    manager.initialize(interactive=interactive,
                       provider_id=provider_id,
                       storage_suffix=storage_suffix,
                       use_existing_disk_name=disk_name,
                       dry_run=dry_run)
    logs.exit_great_success()

@storage.command()
@click.option('--raw', is_flag=True)
@click.option('--provider-id')
@click.option('--storage-suffix')
def credentials(raw, provider_id, storage_suffix):
    manager.print_credentials(raw, provider_id=provider_id, storage_suffix=storage_suffix)



import subprocess
import json
import yaml

from ckan_cloud_operator import logs


class DeisCkanInstanceSolr(object):

    def __init__(self, instance):
        self.instance = instance
        self.solr_spec = self.instance.spec.solrCloudCollection

    def update(self):
        self.instance.annotations.update_status('solr', 'created', lambda: self._update(), force_update=True)

    def delete(self):
        collection_name = self.solr_spec['name']
        logs.warning(f'Not deleting solrcloud collection {collection_name}')
        # from ckan_cloud_operator.providers.solr import manager as solr_manager
        # solr_manager.delete_collection(collection_name)

    def get_replication_factor(self):
        from ckan_cloud_operator.providers.solr import manager as solr_manager
        return solr_manager.get_replication_factor()

    def get_num_shards(self):
        from ckan_cloud_operator.providers.solr import manager as solr_manager
        return solr_manager.get_num_shards()

    def get(self):
        from ckan_cloud_operator.providers.solr import manager as solr_manager
        collection_name = self.instance.spec.solrCloudCollection['name']
        return solr_manager.get_collection_status(collection_name)

    def is_ready(self):
        return self.get().get('ready')

    def _update(self):
        status = self.get()
        if status['ready']:
            schema_name = status['schemaName']
            schema_version = status['schemaVersion']
            logs.info(f'Using existing solr schema: {schema_name} {schema_version}')
        elif 'configName' in self.solr_spec:
            config_name = self.solr_spec['configName']
            from ckan_cloud_operator.providers.solr import manager as solr_manager
            solr_manager.create_collection(status['collection_name'], config_name)
        else:
            raise NotImplementedError(f'Unsupported solr cloud collection spec: {self.solr_spec}')



#### standard provider code ####

# import the correct PROVIDER_SUBMODULE and PROVIDER_ID constants for your provider
from .constants import PROVIDER_ID
from ..constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None, short=False): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix, short=short)
def _get_resource_labels(for_deployment=False): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file)

################################
# custom provider code starts here
#


import tempfile
import traceback
import json
import datetime
import subprocess
import time
from ruamel import yaml
from ckan_cloud_operator import logs
from ckan_cloud_operator import kubectl
from ckan_cloud_operator.drivers.kubectl import rbac as kubectl_rbac_driver
from ckan_cloud_operator.routers import manager as routers_manager
from ckan_cloud_operator.crds import manager as crds_manager
from ckan_cloud_operator.providers.apps.constants import APP_CRD_SINGULAR
from ckan_cloud_operator.drivers.helm import driver as helm_driver


def create(tiller_namespace_name=None, chart_repo=None, chart_version=None, chart_release_name=None,
           values=None, values_filename=None, with_service_account=False, chart_name=None,
           chart_repo_name=None, app_type=None, values_json=None,
           **create_kwargs):
    if values_filename:
        assert not values and not values_json
        with open(values_filename) as f:
            values = yaml.safe_load(f.read())
    elif values:
        assert not values_filename and not values_json
    elif values_json:
        assert not values_filename and not values
        values = json.loads(values_json)
    values = values or {}
    spec = {
        **({'tiller-namespace-name': tiller_namespace_name} if tiller_namespace_name else {}),
        **({'chart-name': chart_name} if chart_name else {}),
        **({'chart-repo': chart_repo} if chart_repo else {}),
        **({'chart-repo-name': chart_repo_name} if chart_repo_name else {}),
        **({'chart-version': chart_version} if chart_version else {}),
        **({'chart-release-name': chart_release_name} if chart_release_name else {}),
        **({'with-service-account': True} if with_service_account else {}),
        **({'app-type': app_type} if app_type else {}),
        **values,
    }
    assert 'values' not in create_kwargs
    from ckan_cloud_operator.providers.apps import manager as apps_manager
    return apps_manager.create('helm', values=spec, **create_kwargs)


def update(instance_id, instance, dry_run=False):
    tiller_namespace_name = _get_tiller_namespace_name(instance_id, instance)
    logs.debug('Updating helm-based instance deployment',
               instance_id=instance_id, tiller_namespace_name=tiller_namespace_name)
    chart_repo_name = instance['spec'].get("chart-repo-name")
    assert chart_repo_name, 'missing spec attribute: chart-repo-name'
    logs.info(chart_repo_name=chart_repo_name)
    chart_repo = instance['spec'].get("chart-repo")
    assert chart_repo or chart_repo_name in ['stable'], 'missing spec attribute: chart-repo'
    logs.info(chart_repo=chart_repo)
    chart_name = instance['spec'].get('chart-name')
    assert chart_name, 'missing spec attribute: chart-name'
    logs.info(chart_name=chart_name)
    chart_version = instance['spec'].get("chart-version", "")
    logs.info(chart_version=chart_version)
    release_name = _get_helm_release_name(instance_id, instance)
    logs.info(release_name=release_name,)
    _pre_update_hook_modify_spec(instance_id, instance, lambda i: i.update(**{
        'release-name': release_name,
        'chart-version': chart_version,
        'chart-name': chart_name,
        'chart-repo': chart_repo,
        'chart-repo-name': chart_repo_name,
    }))
    deploy_kwargs = dict(
        values=instance['spec'].get('values', {}),
        tiller_namespace_name=tiller_namespace_name,
        chart_repo=chart_repo,
        chart_version=chart_version,
        chart_name=chart_name,
        release_name=release_name,
        instance_id=instance_id,
        dry_run=dry_run,
        chart_repo_name=chart_repo_name
    )
    app_type = instance['spec'].get('app-type')
    if app_type:
        _get_app_type_manager(app_type).pre_deploy_hook(instance_id, instance, deploy_kwargs)
    _helm_deploy(**deploy_kwargs)
    if app_type:
        _get_app_type_manager(app_type).post_deploy_hook(instance_id, instance, deploy_kwargs)


def get(instance_id, instance=None):
    res = {
        'ready': None,
        'helm_metadata': {
            'ckan_instance_id': instance_id,
            'namespace': instance_id,
            'status_generated_at': datetime.datetime.now(),
            'status_generated_from': subprocess.check_output(["hostname"]).decode().strip(),
        }
    }
    app_type = instance['spec'].get('app-type')
    if app_type:
        _get_app_type_manager(app_type).get(instance_id, instance, res)
    return res


def delete(instance_id, instance):
    tiller_namespace_name = _get_tiller_namespace_name(instance_id, instance)
    release_name = _get_helm_release_name(instance_id, instance)
    logs.info(tiller_namespace_name=tiller_namespace_name, release_name=release_name)
    errors = []
    try:
        logs.info(f'Deleting helm release {release_name}')
        delete_kwargs=dict(tiller_namespace=tiller_namespace_name, release_name=release_name)
        app_type = instance['spec'].get('app-type')
        if app_type:
            _get_app_type_manager(app_type).pre_delete_hook(instance_id, instance, delete_kwargs)
        helm_driver.delete(**delete_kwargs)
        if app_type:
            _get_app_type_manager(app_type).post_delete_hook(instance_id, instance, delete_kwargs)
    except Exception as e:
        logs.warning(traceback.format_exc())
        errors.append(f'Failed to delete helm release')
    if kubectl.call(f'delete --wait=false namespace {instance_id}') != 0:
        errors.append(f'Failed to delete namespace')
    assert len(errors) == 0, ', '.join(errors)


def get_backend_url(instance_id, instance):
    app_type = instance['spec'].get('app-type')
    if app_type:
        backend_url = instance['spec'].get('backend-url')
        if backend_url:
            try:
                backend_url = backend_url.format(instance_id=instance_id)
            except Exception:
                backend_url = None
        return _get_app_type_manager(app_type).get_backend_url(instance_id, instance, backend_url)
    else:
        return instance['spec'].get('backend-url').format(
            instance_id=instance_id
        )


def pre_update_hook(instance_id, instance, override_spec, skip_route=False, dry_run=False):
    _init_namespace(instance_id, instance, dry_run=dry_run)
    _pre_update_hook_override_spec(override_spec, instance)
    res = {}
    sub_domain, root_domain = _pre_update_hook_route(instance_id, skip_route, instance, res, dry_run=dry_run)
    app_type = instance['spec'].get('app-type')
    logs.info(app_type=app_type)
    if app_type:
        logs.info(f'Running {app_type} app pre_update_hook')
        _get_app_type_manager(app_type).pre_update_hook(
            instance_id, instance, res, sub_domain, root_domain,
            lambda callback: _pre_update_hook_modify_spec(instance_id, instance, callback, dry_run=dry_run)
        )
    return res


def _init_namespace(instance_id, instance, dry_run=False):
    logs.debug('Initializing helm-based instance deployment namespace', namespace=instance_id)
    if kubectl.get('ns', instance_id, required=False):
        logs.info(f'instance namespace already exists ({instance_id})')
    else:
        logs.info(f'creating instance namespace ({instance_id})')
        kubectl.apply(kubectl.get_resource('v1', 'Namespace', instance_id, {}), dry_run=dry_run)
        if instance['spec'].get('with-service-account'):
            service_account_name = instance['spec'].get('service-account-name', f'ckan-{instance_id}-operator')
            logs.info('Creating service account', service_account_name=service_account_name)
            if not dry_run:
                kubectl_rbac_driver.update_service_account(service_account_name, {}, namespace=instance_id)
            role_name = f'{service_account_name}-role'
            logs.debug('Creating role and binding to the service account', role_name=role_name)
            if not dry_run:
                rbac_rules = instance['spec'].get('service-account-rules', [
                    {
                        "apiGroups": [
                            "*"
                        ],
                        "resources": [
                            'secrets', 'pods', 'pods/exec', 'pods/portforward'
                        ],
                        "verbs": [
                            "list", "get", "create"
                        ]
                    }
                ])
                kubectl_rbac_driver.update_role(role_name, {}, rbac_rules, namespace=instance_id)
                kubectl_rbac_driver.update_role_binding(
                    name=f'{service_account_name}-rolebinding',
                    role_name=f'{service_account_name}-role',
                    namespace=instance_id,
                    service_account_name=service_account_name,
                    labels={}
                )


def _pre_update_hook_route(instance_id, skip_route, instance, res, dry_run=False):
    root_domain = routers_manager.get_default_root_domain()
    sub_domain = instance['spec'].get('sub-domain', f'ckan-cloud-app-{instance_id}')
    if not skip_route:
        # full domain to route to the instance
        instance_domain = instance['spec'].get('domain')
        if instance_domain and instance_domain != f'{sub_domain}.{root_domain}':
            logs.warning(f'instance domain was changed from {instance_domain} to {sub_domain}.{root_domain}')
            _pre_update_hook_modify_spec(instance_id, instance,
                                         lambda i: i.update(domain=f'{sub_domain}.{root_domain}'),
                                         dry_run=dry_run)
        # instance is added to router only if this is true, as all routers must use SSL and may use sans SSL too
        with_sans_ssl = instance['spec'].get('withSansSSL')
        if not with_sans_ssl:
            logs.warning(f'forcing with_sans_ssl, even though withSansSSL is disabled')
            _pre_update_hook_modify_spec(instance_id, instance,
                                         lambda i: i.update(withSansSSL=True),
                                         dry_run=dry_run)
        # subdomain to register on the default root domain
        register_subdomain = instance['spec'].get('registerSubdomain')
        if register_subdomain != sub_domain:
            logs.warning(f'instance register sub domain was changed from {register_subdomain} to {sub_domain}')
            _pre_update_hook_modify_spec(instance_id, instance,
                                         lambda i: i.update(registerSubdomain=sub_domain),
                                         dry_run=dry_run)
        res.update(**{'root-domain': root_domain, 'sub-domain': sub_domain})
        site_url = instance['spec'].get('siteUrl')
        if site_url != f'https://{sub_domain}.{root_domain}':
            logs.warning(f'instance siteUrl was changed from {site_url} to https://{sub_domain}.{root_domain}')
            _pre_update_hook_modify_spec(instance_id, instance,
                                         lambda i: i.update(siteUrl=f'https://{sub_domain}.{root_domain}'),
                                         dry_run=dry_run)
    return sub_domain, root_domain


def _pre_update_hook_override_spec(override_spec, instance):
    # applies override spec, but doesn't persist
    if override_spec:
        for k, v in override_spec.items():
            logs.info(f'Applying override spec {k}={v}')
            if k != 'values':
                instance['spec'][k] = v
            else:
                instance['spec'].setdefault('values', {}).update(v)


def _pre_update_hook_modify_spec(instance_id, instance, callback, dry_run=False):
    # applies changes to both the non-persistent spec and persists the changes on latest instance spec
    latest_instance = crds_manager.get(APP_CRD_SINGULAR, name=instance_id, required=True)
    callback(instance['spec'])
    callback(latest_instance['spec'])
    kubectl.apply(latest_instance, dry_run=dry_run)


def _helm_deploy(values, tiller_namespace_name, chart_repo, chart_name, chart_version, release_name, instance_id,
                 dry_run=False, chart_repo_name=None):
    assert chart_repo_name, 'chart-repo-name is required'
    helm_driver.init(tiller_namespace_name)
    time.sleep(10) # wait for tiller pod to be ready
    logs.info(f'Deploying helm chart {chart_repo_name} {chart_repo} {chart_version} {chart_name} to release {release_name} '
              f'(instance_id={instance_id})')
    with tempfile.NamedTemporaryFile('w') as f:
        yaml.dump(values, f, default_flow_style=False)
        f.flush()
        helm_driver.deploy(tiller_namespace=tiller_namespace_name,
                           chart_repo=chart_repo,
                           chart_name=chart_name,
                           chart_version=chart_version,
                           release_name=release_name,
                           values_filename=f.name,
                           namespace=instance_id,
                           dry_run=dry_run,
                           chart_repo_name=chart_repo_name)


def _get_tiller_namespace_name(instance_id, instance):
    return instance['spec'].get('tiller-namespace-name', _get_resource_name('tiller'))


def _get_helm_release_name(instance_id, instance):
    return instance['spec'].get('chart-release-name', _get_resource_name(instance_id, short=False))


def _get_app_type_manager(app_type):
    if app_type == 'provisioning':
        from . import type_provisioning as app_type_manager
    elif app_type == 'jenkins':
        from . import type_jenkins as app_type_manager
    elif app_type == 'elk':
        from . import type_elk as app_type_manager
    else:
        raise NotImplementedError(f'Unknown app type: {app_type}')
    return app_type_manager



#### standard provider code ####

# import the correct PROVIDER_SUBMODULE and PROVIDER_ID constants for your provider
from ckan_cloud_operator.providers.users.gcloud.constants import PROVIDER_ID
from ckan_cloud_operator.providers.users.constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)

################################
# custom provider code starts here
#

from ckan_cloud_operator import kubectl

from ckan_cloud_operator.drivers.kubectl import rbac
from ckan_cloud_operator.crds import manager as crds_manager
from ckan_cloud_operator.providers.users.constants import CRD_SINGULAR
from ckan_cloud_operator.providers.users.manager import get_user_labels, get_role_labels
from ckan_cloud_operator.providers.cluster import manager as cluster_manager


def initialize(interactive=False):
    _set_provider()


def update(user):
    _update_service_account(user)
    _update_role_binding(user)


def get_kubeconfig(user):
    return rbac.get_kubeconfig(
        cluster_name=cluster_manager.get_cluster_name(),
        service_account_name=_get_user_resource_name(user),
        service_account_namespace=_get_user_resource_namespace(user),
        cluster_spec=cluster_manager.get_cluster_kubeconfig_spec()
    )


def _get_name_role(user):
    return user['spec']['name'], user['spec']['role']


def _get_user_resource_name(user):
    return user['metadata']['name']


def _get_role_resource_name(user):
    _, role = _get_name_role(user)
    return crds_manager.get_resource_name(CRD_SINGULAR, f'role-{role}')


def _get_user_role_resource_name(user):
    name, role = _get_name_role(user)
    return crds_manager.get_resource_name(CRD_SINGULAR, f'user-{name}-role-{role}')


def _get_user_resource_namespace(user):
    return user['metadata']['namespace']


def _get_user_resource_labels(user):
    name, role = _get_name_role(user)
    return crds_manager.get_resource_labels(
        CRD_SINGULAR, name,
        extra_label_suffixes=get_user_labels(name, role)
    )


def _get_role_resource_labels(user):
    _, role = _get_name_role(user)
    return crds_manager.get_resource_labels(
        CRD_SINGULAR, role,
        extra_label_suffixes=get_role_labels(role)
    )


def _update_service_account(user):
    rbac.update_service_account(
        service_account_name=_get_user_resource_name(user),
        labels=_get_user_resource_labels(user)
    )


def _update_role_binding(user):
    name, role = _get_name_role(user)
    if role == 'admin':
        rbac.update_admin_cluster_role(
            name=_get_role_resource_name(user),
            labels=_get_role_resource_labels(user),
        )
        rbac.update_cluster_role_binding(
            name=_get_user_role_resource_name(user),
            subject={
                'kind': 'User',
                'name': f'system:serviceaccount:{_get_user_resource_namespace(user)}:{_get_user_resource_name(user)}',
                'apiGroup': 'rbac.authorization.k8s.io'
            },
            cluster_role_name=_get_role_resource_name(user),
            labels=_get_user_resource_labels(user)
        )
    else:
        raise NotImplementedError(f'unsupported role: {role}')



import click
import yaml

from ckan_cloud_operator import logs

from . import manager


@click.group()
def crds():
    pass


@crds.command()
def initialize():
    """Interactive initialization of the crds subsystem"""
    manager.initialize()
    logs.exit_great_success()


@crds.command()
@click.option('--full', is_flag=True)
@click.option('--debug', is_flag=True)
def list_crds(full, debug):
    """List the installed crds"""
    for crd in manager.list_crds(full, debug):
        print(yaml.dump([crd], default_flow_style=False))


@crds.command()
@click.argument('SINGULAR')
@click.option('--full', is_flag=True)
@click.option('--debug', is_flag=True)
def get_crd(singular, full, debug):
    """Get metadata for a custom resource definition"""
    print(yaml.dump(manager.get_crd(singular, full, debug), default_flow_style=False))


@crds.command()
@click.option('--singular')
@click.option('--name')
def list_resources(singular, name):
    for resource in manager.list_resources(singular, name):
        print(yaml.dump([resource], default_flow_style=False))



import click
import time
import tempfile
import traceback
import os
import subprocess

from ckan_cloud_operator.deis_ckan.instance import DeisCkanInstance
from ckan_cloud_operator.infra import CkanInfra

import ckan_cloud_operator.routers.cli
import ckan_cloud_operator.datapushers
from ckan_cloud_operator.gitlab import CkanGitlab
from ckan_cloud_operator import gcloud
import ckan_cloud_operator.storage
from ckan_cloud_operator import logs

from ckan_cloud_operator.providers import cli as providers_cli
from ckan_cloud_operator.providers.db import cli as db_cli
from ckan_cloud_operator.providers.cluster import cli as cluster_cli
from ckan_cloud_operator.providers.users import cli as users_cli
from ckan_cloud_operator.crds import cli as crds_cli
from ckan_cloud_operator.config import cli as config_cli
from ckan_cloud_operator.drivers.postgres import cli as driver_postgres_cli
from ckan_cloud_operator.providers.ckan import cli as ckan_cli
from ckan_cloud_operator.drivers.kubectl import cli as driver_kubectl_cli
from ckan_cloud_operator.providers.storage import cli as storage_cli
from ckan_cloud_operator.providers.solr import cli as solr_cli
from ckan_cloud_operator.drivers.rancher import cli as driver_rancher_cli
from ckan_cloud_operator.drivers.jenkins import cli as driver_jenkins_cli
from ckan_cloud_operator.drivers.helm import cli as driver_helm_cli
from ckan_cloud_operator.providers.apps import cli as apps_cli


CLICK_CLI_MAX_CONTENT_WIDTH = 200


def great_success(**kwargs):
    logs.info('Great Success!', **kwargs)
    exit(0)


@click.group(context_settings={'max_content_width': CLICK_CLI_MAX_CONTENT_WIDTH})
@click.option('--debug', is_flag=True)
def main(debug):
    """Manage, provision and configure CKAN Clouds and related infrastructure"""
    if debug:
        os.environ.setdefault('CKAN_CLOUD_OPERATOR_DEBUG', 'y')
    pass


main.add_command(ckan_cli.ckan)
main.add_command(cluster_cli.cluster)


############################################################################
## Commands Below are not used for now and are making unnecessary noise   ##
## I'm commenting them for now, but not deleting as they might be useful  ##
## In future. Will bring them back as needed                              ##
############################################################################
#
#
# main.add_command(users_cli.users)
# main.add_command(db_cli.db_group, 'db')
# main.add_command(providers_cli.providers_group, 'providers')
# main.add_command(crds_cli.crds)
# main.add_command(config_cli.config)
# main.add_command(storage_cli.storage)
# main.add_command(solr_cli.solr)
# main.add_command(apps _cli.apps)
#
#
# @main.group()
# def drivers():
#     pass
#
#
# drivers.add_command(driver_postgres_cli.postgres)
# drivers.add_command(driver_kubectl_cli.kubectl)
# drivers.add_command(driver_rancher_cli.rancher)
# drivers.add_command(driver_jenkins_cli.jenkins)
# drivers.add_command(driver_helm_cli.helm)
#
#
# @main.command('kubectl')
# @click.argument('ARG', nargs=-1)
# def kubectl_command(arg):
#     from ckan_cloud_operator import kubectl
#     kubectl.check_call(' '.join(arg))
#
#
# @main.command()
# @click.argument('GITLAB_PROJECT_NAME')
# @click.argument('GIT_BRANCH', default='master')
# @click.option('-w', '--wait-ready', is_flag=True)
# def initialize_gitlab(gitlab_project_name, git_branch, wait_ready):
#     """Initialize the gitlab integration
#
#     Example:
#
#         ckan-cloud-operator initialize-gitlab repo/project
#     """
#     ckan_gitlab = CkanGitlab()
#     ckan_gitlab.initialize(gitlab_project_name, git_branch)
#     if wait_ready and not ckan_gitlab.is_ready(gitlab_project_name):
#         logs.info(f'Waiting for GitLab project {gitlab_project_name} to be ready...')
#         while not ckan_gitlab.is_ready(gitlab_project_name):
#             time.sleep(5)
#     great_success()
#
#
# @main.command()
# def activate_gcloud_auth():
#     """Authenticate with gcloud CLI using the ckan-cloud-operator credentials"""
#     from ckan_cloud_operator.providers.cluster.gcloud import manager as gcloud_manager
#     gcloud_manager.activate_auth()
#
#
# @main.command()
# def bash_completion():
#     """Return bash completion script which should be eval'd"""
#     subprocess.check_call('echo "$(_CKAN_CLOUD_OPERATOR_COMPLETE=source ckan-cloud-operator)"', shell=True)
#     print('# ')
#     print('# To enable Bash completion, use the following command:')
#     print('# eval "$(ckan-cloud-operator bash-completion)"')
#
#
# @main.group()
# def users():
#     """Manage ckan-cloud-operator users"""
#     pass
#
# ckan_cloud_operator.providers.users.add_cli_commands(click, users, great_success)
#
#
# @main.group()
# def ckan_infra():
#     """Manage the centralized infrastructure"""
#     pass
#
#
# CkanInfra.add_cli_commands(click, ckan_infra, great_success)
#
#
# @main.group()
# def deis_instance():
#     """Manage Deis CKAN instance resources"""
#     pass
#
#
# DeisCkanInstance.add_cli_commands(click, deis_instance, great_success)
#
#
# @main.group()
# def routers():
#     """Manage CKAN Cloud routers"""
#     pass
#
#
# ckan_cloud_operator.routers.cli.add_cli_commands(click, routers, great_success)
#
#
# @main.group()
# def datapushers():
#     """Manage centralized CKAN DataPushers"""
#     pass
#
#
# ckan_cloud_operator.datapushers.add_cli_commands(click, datapushers, great_success)
#
#
# @main.command()
# def test():
#     """Run unittest suite and report coverage"""
#     subprocess.check_call('coverage run -m unittest discover', shell=True)
#     subprocess.check_call('coverage report', shell=True)


if __name__ == '__main__':
    main()



import datetime

from ckan_cloud_operator import kubectl

from ckan_cloud_operator.labels import manager as labels_manager


def get_global_annotations(with_timestamp=True):
    label_prefix = labels_manager.get_label_prefix()
    return {
        f'{label_prefix}/operator-timestamp': str(datetime.datetime.now())
    } if with_timestamp else {}


def set_status(resource, prefix, status):
    _annotate(resource, f'{prefix}-{status}=true')


def get_status(resource, prefix, status=None):
    if status:
        return bool(_get_annotation(resource, f'{prefix}-{status}'))
    else:
        label_prefix = labels_manager.get_label_prefix()
        prefix = f'{label_prefix}/{prefix}'
        statuses = set()
        for k, v in resource['metadata'].get('annotations', {}).items():
            if k.startswith(prefix):
                statuses.add(k.replace(prefix, ''))
        return list(statuses)


def _annotate(resource, *annotations, overwrite=True):
    label_prefix = labels_manager.get_label_prefix()
    kind = resource['kind']
    namespace = resource['metadata']['namespace']
    name = resource['metadata']['name']
    cmd = f'annotate {kind} {name}'
    for annotation in annotations:
        cmd += f' {label_prefix}/{annotation}'
    if overwrite:
        cmd += ' --overwrite'
    kubectl.check_call(cmd, namespace)


def _get_annotation(resource, annotation, default=None):
    label_prefix = labels_manager.get_label_prefix()
    return resource['metadata'].get('annotations', {}).get(f'{label_prefix}/{annotation}', default)



#### standard provider code ####

# import the correct PROVIDER_SUBMODULE and PROVIDER_ID constants for your provider
from .constants import PROVIDER_ID
from ..constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file)

################################
# custom provider code starts here
#

import binascii
import json
import os
import subprocess
import yaml

from ckan_cloud_operator import logs
from ckan_cloud_operator import kubectl


def initialize(interactive=False):
    _set_provider()
    if interactive:
        print('\nEnter the Resource Group name\n')
        _config_interactive_set({'azure-rg': None}, is_secret=False)
        print('\nEnter the location of the Kubernets cluster is hosted on [westus2]\n')
        _config_interactive_set({'azure-default-location': None}, is_secret=False)
        print('\nEnter the name of your cluster\n')
        _config_interactive_set({'azure-cluster-name': None}, is_secret=False)
        print('\nEnter the Subscribtion ID\n')
        _config_interactive_set({'azure-subscribtion-id': None}, is_secret=True)
        print('\nEnter the Tenant ID\n')
        _config_interactive_set({'azure-tenant-id': None}, is_secret=True)
        print('\nEnter the Service Principal ID\n')
        _config_interactive_set({'azure-client-id': None}, is_secret=True)
        print('\nEnter the Service Principal Secret\n')
        _config_interactive_set({'azure-client-secret': None}, is_secret=True)

        _create_storage_classes()
    else:
        logs.info('Skipping initial cluster set up as `--interactive` flag was not set')
    _create_storage_classes()


def get_info(debug=False):
    cluster_name = _config_get('cluster-name')
    data = yaml.load(az_check_output(f'container clusters describe {cluster_name}'))
    if debug:
        return data
    else:
        return {
            'name': data['name'],
            'status': data['status'],
            'zone': data['zone'],
            'locations': data['locations'],
            'endpoint': data['endpoint'],
            'nodePools': [
                {
                    'name': pool['name'],
                    'status': pool['status'],
                    'version': pool['version'],
                    'config': {
                        'diskSizeGb': pool['config']['diskSizeGb'],
                        'machineType': pool['config']['machineType'],
                    },
                } for pool in data['nodePools']
            ],
            'createTime': data['createTime'],
            'currentMasterVersion': data['currentMasterVersion'],
            'currentNodeCount': data['currentNodeCount'],
        }


def get_name():
    return _config_get('cluster-name')

def get_azure_credentials():
    return {
        'azure-client-id': _config_get('azure-client-id', is_secret=True),
        'azure-client-secret': _config_get('azure-client-secret', is_secret=True),
        'azure-subscribtion-id': _config_get('azure-subscribtion-id', is_secret=True),
        'azure-tenant-id': _config_get('azure-tenant-id', is_secret=True),
        'azure-resource-group': _config_get('azure-rg')
    }

def _create_storage_classes():
    kubectl.apply({
        'apiVersion': 'storage.k8s.io/v1',
        'kind': 'StorageClass',
        'metadata': {
            'name': 'cca-ckan',
        },
        'parameters': {
            'skuName': 'Standard_LRS',
            'location': _config_get('azure-default-location')
        },
        'provisioner': 'kubernetes.io/azure-disk',
        'reclaimPolicy': 'Delete',
        'volumeBindingMode': 'Immediate'
    })
    kubectl.apply({
        'apiVersion': 'storage.k8s.io/v1',
        'kind': 'StorageClass',
        'metadata': {
            'name': 'cca-storage',
        },
        'provisioner': 'kubernetes.io/azure-disk',
        'volumeBindingMode': 'Immediate',
        'parameters': {
            'skuName': 'Standard_LRS',
            'location': _config_get('azure-default-location')
        }
    })

def get_cluster_kubeconfig_spec():
    cluster_name = _config_get('cluster-name')
    cluster = yaml.load(az_check_output(f'container clusters describe {cluster_name}'))
    return {
        "server": 'https://' + cluster['endpoint'],
        "certificate-authority-data": cluster['masterAuth']['clusterCaCertificate']
    }


def get_project_zone():
    return _config_get('project-id'), _config_get('cluster-compute-zone')


def create_volume(disk_size_gb, labels, use_existing_disk_name=None, zone=0):
    rg = _config_get('azure-rg')

    location = zone or _config_get('azure-default-location')

    disk_id = use_existing_disk_name or 'cc' + _generate_password(12)
    if use_existing_disk_name:
        logs.info(f'using existing persistent disk {disk_id}')
    else:
        logs.info(f'creating persistent disk {disk_id} with size {disk_size_gb}GB')
        _, zone = get_project_zone()
        labels = ','.join([
            '{}={}'.format(k.replace('/', '_'), v.replace('/', '_')) for k, v in labels.items()
        ])

    kubectl.apply({
        "kind": "PersistentVolumeClaim",
        "apiVersion": "v1",
        "metadata": {"name": disk_id,"namespace": "ckan-cloud"},
        "spec": {
            "accessModes": ["ReadWriteOnce"],
            "resources": {
                "requests": {
                    "storage": f'{disk_size_gb}G'
                }
            },
            "storageClassName": "cca-ckan"
        }
    })
    return {'persistentVolumeClaim': {'claimName': disk_id}}


def _generate_password(l):
    return binascii.hexlify(os.urandom(l)).decode()


def az_check_output(cmd):
    return subprocess.check_output(f'az {cmd}', shell=True)


def create_dns_record(sub_domain, root_domain, load_balancer_ip_or_hostname):
    logs.info('updating Azure DNS record', sub_domain=sub_domain, root_domain=root_domain, load_balancer_hostname=load_balancer_ip_or_hostname)
    resource_group = _config_get('azure-rg')
    # az network dns record-set a show exits with non 0 if DNS record name not found
    try:
        # Check if exists and do nothing...
        cmd = f'network dns record-set a show  -g {resource_group} -z {root_domain} -n {sub_domain}'
        az_check_output(cmd)
    except:
        # Create if does not
        cmd = f'network dns record-set a add-record  -g {resource_group} -z {root_domain} -n {sub_domain} -a {load_balancer_ip_or_hostname}'
        az_check_output(cmd)



import traceback

from ckan_cloud_operator import logs

import ckan_cloud_operator.routers.routes.manager as routes_manager


def _get_base_config(**kwargs):
    logs.info('Generating base Traefik configuration', **kwargs)
    return dict({
        'debug': False,
        'defaultEntryPoints': ['http'],
        'entryPoints': {
            'http': {
                'address': ':80'
            }
        },
        'ping': {
            'entryPoint': 'http'
        },
        'file': {},
        'frontends': {},
        'backends': {},
    }, **kwargs)


def _get_acme_domains(domains, wildcard_ssl_domain=None, external_domains=False):
    for root_domain, sub_domains in domains.items():
        if external_domains:
            for sub_domain in sub_domains:
                yield {'main': f'{sub_domain}.{root_domain}'}
        elif root_domain == wildcard_ssl_domain:
            yield {
                'main': f'*.{root_domain}',
            }
        else:
            yield {
                'main': root_domain,
                'sans': [f'{sub_domain}.{root_domain}' for sub_domain in sub_domains]
            }


def _add_letsencrypt(dns_provider, config, letsencrypt_cloudflare_email, domains,
                     wildcard_ssl_domain=None, external_domains=False):
    logs.info('Adding Letsencrypt acme Traefik configuration', dns_provider=dns_provider,
              letsencrypt_cloudflare_email=letsencrypt_cloudflare_email, domains=domains,
              wildcard_ssl_domain=wildcard_ssl_domain, external_domains=external_domains)
    assert dns_provider in ['route53', 'cloudflare', 'azure']
    config['defaultEntryPoints'].append('https')
    config['entryPoints']['https'] = {
        'address': ':443',
        'tls': {}
    }
    config['acme'] = {
        'email': letsencrypt_cloudflare_email,
        'storage': '/traefik-acme/acme.json',
        'entryPoint': 'https',
        **(
            {
                'tlsChallenge': {}
            } if external_domains else {
                'dnsChallenge': {
                    'provider': dns_provider
                }
            }
        ),
        'domains': list(_get_acme_domains(domains, wildcard_ssl_domain=wildcard_ssl_domain,
                                          external_domains=external_domains))
    }


def _add_route(config, domains, route, enable_ssl_redirect):
    route_name = routes_manager.get_name(route)
    logs.info(f'adding route to traefik config: {route_name}')
    logs.debug_verbose(config=config, domains=domains, route=route, enable_ssl_redirect=enable_ssl_redirect)
    backend_url = routes_manager.get_backend_url(route)
    frontend_hostname = routes_manager.get_frontend_hostname(route)
    print(f'F/B = {frontend_hostname} {backend_url}')
    root_domain, sub_domain = routes_manager.get_domain_parts(route)
    domains.setdefault(root_domain, []).append(sub_domain)
    if route['spec'].get('extra-no-dns-subdomains'):
        extra_hostnames = ',' + ','.join([f'{s}.{root_domain}' for s in route['spec']['extra-no-dns-subdomains']])
    else:
        extra_hostnames = ''
    logs.debug_verbose(route_name=route_name, backend_url=backend_url, frontend_hostname=frontend_hostname, root_domain=root_domain,
                       sub_domain=sub_domain, domains=domains, extra_hostnames=extra_hostnames)
    if backend_url:
        config['backends'][route_name] = {
            'servers': {
                'server1': {
                    'url': backend_url
                }
            }
        }
        config['frontends'][route_name] = {
            'backend': route_name,
            'passHostHeader': True,
            'headers': {
                'SSLRedirect': bool(enable_ssl_redirect)
            },
            'routes': {
                'route1': {
                    'rule': f'Host:{frontend_hostname}{extra_hostnames}'
                }
            },
            **({
                'auth': {
                    'basic': {
                        'usersFile': '/httpauth-' + route['spec']['httpauth-secret'] + '/.htpasswd'
                    }
                }
            } if route['spec'].get('httpauth-secret') else {}),
        }


def get(routes, letsencrypt_cloudflare_email, enable_access_log=False, wildcard_ssl_domain=None, external_domains=False,
        dns_provider=None, force=False):
    if not dns_provider:
        dns_provider = 'cloudflare'
    logs.info('Generating traefik configuration', routes_len=len(routes) if routes else 0,
              letsencrypt_cloudflare_email=letsencrypt_cloudflare_email, enable_access_log=enable_access_log,
              wildcard_ssl_domain=wildcard_ssl_domain, external_domains=external_domains)
    config = _get_base_config(
        **(
            {
                'accessLog': {
                    "format": "json",
                    "fields": {
                        'defaultMode': "keep"
                    }
                },
            }
            if enable_access_log else {}
        )
    )
    domains = {}
    if dns_provider == 'cloudflare' and letsencrypt_cloudflare_email:
        enable_ssl_redirect = True
    elif dns_provider == 'route53':
        enable_ssl_redirect = True
    elif dns_provider == 'azure':
        enable_ssl_redirect = True
    else:
        enable_ssl_redirect = False
    logs.info(enable_ssl_redirect=enable_ssl_redirect)
    logs.info('Adding routes')
    i = 0
    errors = 0
    for route in routes:
        try:
            _add_route(config, domains, route, enable_ssl_redirect)
            i += 1
        except Exception as e:
            if force:
                logs.error(traceback.format_exc())
                logs.error(str(e))
                errors += 1
            else:
                raise
    logs.info(f'Added {i} routes')
    if errors > 0:
        logs.warning(f'Encountered {errors} errors')
    if (
        (dns_provider == 'cloudflare' and letsencrypt_cloudflare_email)
        or (dns_provider == 'route53')
        or (dns_provider == 'azure')
    ):
        _add_letsencrypt(dns_provider, config, letsencrypt_cloudflare_email, domains,
                         wildcard_ssl_domain=wildcard_ssl_domain, external_domains=external_domains)
    else:
        logs.info('No valid dns_provider, will not setup SSL', dns_provider=dns_provider)
    return config



import subprocess
import yaml
import traceback
import time
import json

from ckan_cloud_operator import kubectl
from ckan_cloud_operator.infra import CkanInfra
from ckan_cloud_operator.deis_ckan.ckan import DeisCkanInstanceCKAN
from ckan_cloud_operator.deis_ckan.annotations import DeisCkanInstanceAnnotations
from ckan_cloud_operator.deis_ckan.db import DeisCkanInstanceDb
from ckan_cloud_operator.deis_ckan.deployment import DeisCkanInstanceDeployment
from ckan_cloud_operator.deis_ckan.envvars import DeisCkanInstanceEnvvars
from ckan_cloud_operator.deis_ckan.namespace import DeisCkanInstanceNamespace
from ckan_cloud_operator.deis_ckan.registry import DeisCkanInstanceRegistry
from ckan_cloud_operator.deis_ckan.solr import DeisCkanInstanceSolr
from ckan_cloud_operator.deis_ckan.spec import DeisCkanInstanceSpec
from ckan_cloud_operator.deis_ckan.storage import DeisCkanInstanceStorage
from ckan_cloud_operator.monitoring.uptime import DeisCkanInstanceUptime
from ckan_cloud_operator.routers import manager as routers_manager
from ckan_cloud_operator import logs
from ckan_cloud_operator.providers.ckan import manager as ckan_manager
from ckan_cloud_operator.providers.ckan.db import migration as ckan_db_migration_manager


class DeisCkanInstance(object):
    """Root object for Deis CKAN instances"""

    @classmethod
    def add_cli_commands(cls, click, command_group, great_success):

        @command_group.command('list')
        @click.option('-f', '--full', is_flag=True)
        @click.option('-q', '--quick', is_flag=True)
        def deis_instance_list(full, quick):
            """List the Deis instances"""
            cls.list(full, quick)

        @command_group.command('get')
        @click.argument('INSTANCE_ID')
        @click.argument('ATTR', required=False)
        def deis_instance_get(instance_id, attr):
            """Get detailed information about an instance, optionally returning only a single get attribute

            Example: ckan-cloud-operator get <INSTANCE_ID> deployment
            """
            print(yaml.dump(cls(instance_id).get(attr), default_flow_style=False))

        @command_group.command('edit')
        @click.argument('INSTANCE_ID')
        @click.argument('EDITOR', default='nano')
        def deis_instance_edit(instance_id, editor):
            """Launch an editor to modify and update an instance"""
            instance_kind = ckan_manager.instance_kind()
            subprocess.call(f'EDITOR={editor} kubectl -n ckan-cloud edit {instance_kind}/{instance_id}', shell=True)
            cls(instance_id).update()
            great_success()

        @command_group.command('update')
        @click.argument('INSTANCE_ID')
        @click.argument('OVERRIDE_SPEC_JSON', required=False)
        @click.option('--persist-overrides', is_flag=True)
        @click.option('--wait-ready', is_flag=True)
        @click.option('--skip-deployment', is_flag=True)
        def deis_instance_update(instance_id, override_spec_json, persist_overrides, wait_ready, skip_deployment):
            """Update an instance to the latest resource spec, optionally applying the given json override to the resource spec

            Examples:

            ckan-cloud-operator update <INSTANCE_ID> '{"envvars":{"CKAN_SITE_URL":"http://localhost:5000"}}' --wait-ready

            ckan-cloud-operator update <INSTANCE_ID> '{"flags":{"skipDbPermissions":false}}' --persist-overrides
            """
            override_spec = json.loads(override_spec_json) if override_spec_json else None
            cls(instance_id, override_spec=override_spec, persist_overrides=persist_overrides).update(
                wait_ready=wait_ready, skip_deployment=skip_deployment
            )
            great_success()

        @command_group.command('delete')
        @click.argument('INSTANCE_ID', nargs=-1)
        @click.option('--force', is_flag=True)
        def deis_instance_delete(instance_id, force):
            """Permanently delete the instances and all related infrastructure"""
            failed_instance_ids = []
            for id in instance_id:
                try:
                    cls(id).delete(force)
                except Exception:
                    traceback.print_exc()
                    failed_instance_ids.append(id)
            if len(failed_instance_ids) > 0:
                logs.critical(f'Instance deletion failed for the following instance ids: {failed_instance_ids}')
                if not force:
                    logs.exit_catastrophic_failure()
            great_success()

        #### deis-instance create

        @command_group.group('create')
        def deis_instance_create():
            """Create and update an instance"""
            pass

        @deis_instance_create.command('from-gitlab')
        @click.argument('GITLAB_REPO_NAME')
        @click.argument('SOLR_CONFIG_NAME')
        @click.argument('NEW_INSTANCE_ID')
        @click.option('--no-db-proxy', is_flag=True)
        @click.option('--from-db-backups')
        @click.option('--storage-path')
        @click.option('--solr-collection')
        @click.option('--rerun', is_flag=True)
        @click.option('--force', is_flag=True)
        @click.option('--recreate-dbs', is_flag=True)
        @click.option('--db-prefix')
        @click.option('--use-private-gitlab-repo', is_flag=True)
        def deis_instance_create_from_gitlab(gitlab_repo_name, solr_config_name, new_instance_id, no_db_proxy,
                                             from_db_backups, storage_path, solr_collection, rerun,
                                             force, recreate_dbs, db_prefix,
                                             use_private_gitlab_repo):
            """Create and update a new instance from a GitLab repo containing Dockerfile and .env

            Example: ckan-cloud-operator deis-instance create from-gitlab viderum/cloud-demo2 ckan_27_default <NEW_INSTANCE_ID>
            """
            cls.create('from-gitlab', gitlab_repo_name, solr_config_name, new_instance_id,
                       no_db_proxy=no_db_proxy, storage_path=storage_path,
                       from_db_backups=from_db_backups, solr_collection=solr_collection,
                       rerun=rerun, force=force, recreate_dbs=recreate_dbs,
                       db_prefix=db_prefix,
                       use_private_gitlab_repo=use_private_gitlab_repo).update()
            great_success()

        #### deis-instance ckan

        @command_group.group('ckan')
        def deis_instance_ckan():
            """Manage a running CKAN instance"""
            pass

        @deis_instance_ckan.command('paster')
        @click.argument('INSTANCE_ID')
        @click.argument('PASTER_ARGS', nargs=-1)
        def deis_instance_ckan_paster(instance_id, paster_args):
            """Run CKAN Paster commands

            Run without PASTER_ARGS to get the available paster commands from the server

            Examples:

              ckan-cloud-operator deis-instance ckan-paster <INSTANCE_ID> sysadmin add admin name=admin email=admin@ckan

              ckan-cloud-operator deis-instance ckan-paster <INSTANCE_ID> search-index rebuild
            """
            cls(instance_id).ckan.run('paster', *paster_args)

        @deis_instance_ckan.command('port-forward')
        @click.argument('INSTANCE_ID')
        @click.argument('PORT', default='5000')
        def deis_instance_port_forward(instance_id, port):
            """Start a port-forward to the CKAN instance pod"""
            cls(instance_id).ckan.run('port-forward', port)

        @deis_instance_ckan.command('exec')
        @click.argument('INSTANCE_ID')
        @click.argument('KUBECTL_EXEC_ARGS', nargs=-1)
        def deis_instance_ckan_exec(instance_id, kubectl_exec_args):
            """Run kubectl exec on the first CKAN instance pod"""
            cls(instance_id).ckan.run('exec', *kubectl_exec_args)

        @deis_instance_ckan.command('logs')
        @click.argument('INSTANCE_ID')
        @click.argument('KUBECTL_LOGS_ARGS', nargs=-1)
        def deis_instance_ckan_logs(instance_id, kubectl_logs_args):
            """Run kubectl logs on the first CKAN instance pod"""
            cls(instance_id).ckan.run('logs', *kubectl_logs_args)

    def __init__(self, id, values=None, override_spec=None, persist_overrides=False):
        """
        :param id: Deis CKAN instance id
        :param values: dict containing the relevant DeisCkanInstance Kubernetes object
        :param override_spec: optional dict containing temporary override values for the spec
                              these values are not persistent
        :param persist_overrides: if true the override spec will be persisted

        The DeisCkanInstance object is lazy-loaded, when the object spec is required
        these values will be used to fetch the latest spec
        """
        self.id = id
        self._values = values
        self._override_spec = override_spec
        self._persist_overrides = persist_overrides
        self._override_flags = self._override_spec.pop('flags') if self._override_spec and 'flags' in self._override_spec else None

    @property
    def values(self):
        """Initialize the raw Kubernetes resource object dict"""
        values = getattr(self, '_values', None)
        if not values:
            self._values = values = kubectl.get(f'{self.kind} {self.id}')
        return values

    @property
    def kind(self):
        return ckan_manager.instance_kind()

    @property
    def spec(self):
        """Initialize the spec object, fetch values from kubernetes if not provided

        :return: DeisCkanInstanceSpec
        """
        if not getattr(self, '_spec', None):
            self._spec = DeisCkanInstanceSpec(self.values['spec'], self._override_spec)
            if self._persist_overrides and self._spec.num_applied_overrides > 0:
                logs.info('persisting overrides')
                logs.debug(f'saving spec for instance id {self.id}: {self._spec.spec}')
                instance = kubectl.get(f'{self.kind} {self.id}')
                instance['spec'] = self._spec.spec
                logs.debug_yaml_dump(instance)
                kubectl.apply(instance)
        return self._spec

    @property
    def ckan_infra(self):
        """Initialize the infra object, fetching values from Kubernetes secrets

        :return: CkanInfra
        """
        ckan_infra = getattr(self, '_ckan_infra', None)
        if not ckan_infra:
            self._ckan_infra = ckan_infra = CkanInfra(required=False)
        return ckan_infra

    @property
    def annotations(self):
        """Initialize the annotations object for storing instance metadata

        :return: DeisCkanInstanceAnnotations
        """
        annotations = getattr(self, '_annotations', None)
        if not annotations:
            self._annotations = annotations = DeisCkanInstanceAnnotations(self, self._override_flags, self._persist_overrides)
        return annotations

    @property
    def ckan(self):
        """Initializes the ckan object for managing running CKAN instances

        :return: DeisCkanInstanceCKAN
        """
        ckan = getattr(self, '_ckan', None)
        if not ckan:
            self._ckan = ckan = DeisCkanInstanceCKAN(self)
        return ckan

    def update(self, wait_ready=False, skip_solr=False, skip_deployment=False):
        """Ensure the instance is updated to latest spec"""
        old_deployment = kubectl.get(f'deployment {self.id}', required=False, namespace=self.id)
        if old_deployment:
            old_deployment_generation = old_deployment.get('metadata', {}).get('generation')
        else:
            old_deployment_generation = None
        if old_deployment_generation:
            expected_new_deployment_generation = old_deployment_generation + 1
        else:
            expected_new_deployment_generation = 1
        print(f'old deployment generation = {old_deployment_generation}')
        DeisCkanInstanceNamespace(self).update()
        DeisCkanInstanceDb(self, 'db').update()
        DeisCkanInstanceDb(self, 'datastore').update()
        if not skip_solr:
            DeisCkanInstanceSolr(self).update()
        DeisCkanInstanceStorage(self).update()
        DeisCkanInstanceRegistry(self).update()
        envvars = DeisCkanInstanceEnvvars(self)
        envvars.update()
        if not skip_deployment:
            DeisCkanInstanceDeployment(self).update()
            while True:
                time.sleep(.2)
                new_deployment = kubectl.get(f'deployment {self.id}', required=False, namespace=self.id)
                if not new_deployment: continue
                new_deployment_generation = new_deployment.get('metadata', {}).get('generation')
                if not new_deployment_generation: continue
                if new_deployment_generation == old_deployment_generation: continue
                if new_deployment_generation != expected_new_deployment_generation:
                    raise Exception(f'Invalid generation: {new_deployment_generation} '
                                    f'(expected: {expected_new_deployment_generation}')
                print(f'new deployment generation: {new_deployment_generation}')
                break
            if wait_ready:
                print('Waiting for ready status')
                time.sleep(3)
                while True:
                    data = self.get()
                    if data.get('ready'):
                        print(yaml.dump(data, default_flow_style=False))
                        break
                    else:
                        print(yaml.dump(
                            {
                                k: v for k, v in data.items()
                                if (k not in ['ready'] and type(v) == dict and not v.get('ready')) or k == 'namespace'
                            },
                            default_flow_style=False)
                        )
                        time.sleep(2)
        self.ckan.update()
        try:
            DeisCkanInstanceDb(self, 'datastore').set_datastore_readonly_permissions()
        except Exception:
            logs.warning('Setting datastore permissions failed, continuing anyway')
        # Create/Update uptime monitoring after everything else is ready
        DeisCkanInstanceUptime(self).update(envvars.site_url)

    def delete(self, force=False, wait_deleted=False):
        """
        Can run delete multiple time until successful deletion of all components.
        Uses Kubernetes finalizers to ensure deletion is complete before applying the deletion.
        """
        print(f'Deleting {self.kind} {self.id}')
        try:
            assert self.spec
            has_spec = True
        except Exception:
            has_spec = False
        # this updates deletion timestamp but doesn't delete the object until all finalizers are removed
        subprocess.call(f'kubectl -n ckan-cloud delete --wait=false {self.kind} {self.id}', shell=True)
        num_exceptions = 0
        if has_spec:
            for delete_id, delete_code in {
                'deployment': lambda: DeisCkanInstanceDeployment(self).delete(),
                'envvars': lambda: DeisCkanInstanceEnvvars(self).delete(),
                'registry': lambda: DeisCkanInstanceRegistry(self).delete(),
                'solr': lambda: DeisCkanInstanceSolr(self).delete(),
                'storage': lambda: DeisCkanInstanceStorage(self).delete(),
                'namespace': lambda: DeisCkanInstanceNamespace(self).delete(),
                'envvars-secret': lambda: kubectl.check_call(f'delete --ignore-not-found secret/{self.id}-envvars'),
                'routes': lambda: routers_manager.delete_routes(deis_instance_id=self.id),
                'uptime-monitoring': lambda: DeisCkanInstanceUptime(self).delete(self.id)
            }.items():
                try:
                    delete_code()
                except Exception as e:
                    logs.critical(f'deletion failed for instance {self.id}, submodule: {delete_id}')
                    num_exceptions += 1
        else:
            try:
                routers_manager.delete_routes(deis_instance_id=self.id)
            except Exception as e:
                logs.critical(f'deletion failed for instance {self.id}, submodule: routes')
                num_exceptions += 1
            num_exceptions += 1
        if num_exceptions != 0 and not force:
            raise Exception('instance was not deleted, run with --force to force deletion with risk of remaining infra')
        else:
            print(f'Removing finalizers from {self.kind} {self.id}')
            try:
                subprocess.check_call(
                    f'kubectl -n ckan-cloud patch {self.kind} {self.id} -p \'{{"metadata":{{"finalizers":[]}}}}\' --type=merge',
                    shell=True
                )
            except Exception:
                logs.critical(f'failed to remove finalizers: {self.id}')
                num_exceptions += 1
                if not force:
                    raise
        if wait_deleted and has_spec:
            logs.info('Waiting 30 seconds for instance to be deleted...')
            time.sleep(30)

    def kubectl(self, cmd, check_output=False):
        if check_output:
            return subprocess.check_output(f'kubectl -n {self.id} {cmd}', shell=True)
        else:
            subprocess.check_call(f'kubectl -n {self.id} {cmd}', shell=True)

    def get(self, attr=None, exclude_attr=None):
        """Get detailed information about the instance and related components"""
        gets = {
            'annotations': lambda: DeisCkanInstanceAnnotations(self).get(),
            'db': lambda: DeisCkanInstanceDb(self, 'db').get(),
            'datastore': lambda: DeisCkanInstanceDb(self, 'datastore').get(),
            'deployment': lambda: DeisCkanInstanceDeployment(self).get(),
            'envvars': lambda: DeisCkanInstanceEnvvars(self).get(),
            'namespace': lambda: DeisCkanInstanceNamespace(self).get(),
            'registry': lambda: DeisCkanInstanceRegistry(self).get(),
            'solr': lambda: DeisCkanInstanceSolr(self).get(),
            'storage': lambda: DeisCkanInstanceStorage(self).get(),
        }
        if exclude_attr:
            gets = {k: v for k, v in gets.items() if k not in exclude_attr}
        if attr:
            return gets[attr]()
        else:
            ret = {'ready': True}
            for k, v in gets.items():
                ret[k] = v()
                if type(ret[k]) == dict and not ret[k].get('ready'):
                    ret['ready'] = False
            ret['id'] = self.id
            return ret

    @classmethod
    def list(cls, full=False, quick=False, return_list=False):
        res = []
        data = kubectl.get(ckan_manager.instance_kind(), required=False)
        if not data: data = {'items': []}
        for item in data['items']:
            if quick:
                data = {
                    'id': item['metadata']['name'],
                    'ready': None
                }
                if full:
                    data['item'] = item
            else:
                try:
                    instance = DeisCkanInstance(item['metadata']['name'], values=item)
                    data = instance.get()
                    if not full:
                        data = {'id': instance.id, 'ready': data['ready']}
                except Exception:
                    data = {'id': item['metadata']['name'], 'ready': False, 'error': traceback.format_exc()}
            if return_list:
                res.append(data)
            else:
                print(yaml.dump([data], default_flow_style=False))
        if return_list:
            return res


    @classmethod
    def create(cls, *args, **kwargs):
        create_type = args[0]
        instance_id = args[-1]
        from ckan_cloud_operator.providers.db.manager import get_default_db_prefix
        db_prefix = kwargs['db_prefix'] if kwargs.get('db_prefix') else get_default_db_prefix()
        if create_type == 'from-gitlab':
            gitlab_repo = args[1]
            solr_config = args[2]
            db_name = instance_id
            datastore_name = f'{instance_id}-datastore'
            storage_path = kwargs.get('storage_path') or f'/ckan/{instance_id}'
            from_db_backups = kwargs.get('from_db_backups')
            logs.info(f'Creating Deis CKAN instance {instance_id}', gitlab_repo=gitlab_repo, solr_config=solr_config,
                      db_name=db_name, datastore_name=datastore_name, storage_path=storage_path,
                      from_db_backups=from_db_backups)

            if kwargs.get('use_private_gitlab_repo'):
                deploy_token_server = input('Gitlab registry url [default: registry.gitlab.com]: ') or 'registry.gitlab.com'
                deploy_token_username = input('Gitlab deploy token username: ')
                deploy_token_password = input('Gitlab deploy token password: ')
                kubectl.call('delete secret private-gitlab-registry', namespace=instance_id)
                kubectl.call(f'create secret docker-registry private-gitlab-registry --docker-server={deploy_token_server} --docker-username={deploy_token_username} --docker-password={deploy_token_password}', namespace=instance_id)

            if from_db_backups:
                db_import_url, datastore_import_url = from_db_backups.split(',')
                migration_name = None
                success = False
                for event in ckan_db_migration_manager.migrate_deis_dbs(None, db_name, datastore_name,
                                                                        db_import_url=db_import_url,
                                                                        datastore_import_url=datastore_import_url,
                                                                        rerun=kwargs.get('rerun'),
                                                                        force=kwargs.get('force'),
                                                                        recreate_dbs=kwargs.get('recreate_dbs'),
                                                                        db_prefix=db_prefix):
                    migration_name = ckan_db_migration_manager.get_event_migration_created_name(event) or migration_name
                    success = ckan_db_migration_manager.print_event_exit_on_complete(
                        event,
                        f'DBs import {from_db_backups} -> {db_name}, {datastore_name}',
                        soft_exit=True
                    )
                    if success is not None:
                        break
                assert success, f'Invalid DB migration success value ({success})'
            else:
                migration_name = None
            spec = {
                'ckanPodSpec': {},
                'ckanContainerSpec': {'imageFromGitlab': gitlab_repo},
                'envvars': {'fromGitlab': gitlab_repo},
                'solrCloudCollection': {
                    'name': kwargs.get('solr_collection') or instance_id,
                    'configName': solr_config
                },
                'db': {
                    'name': db_name,
                    **({'fromDbMigration': migration_name} if migration_name else {}),
                    **({'dbPrefix': db_prefix} if db_prefix else {})
                },
                'datastore': {
                    'name': datastore_name,
                    **({'fromDbMigration': migration_name} if migration_name else {}),
                    **({'dbPrefix': db_prefix} if db_prefix else {})
                },
                'storage': {
                    'path': storage_path,
                }
            }
            if kwargs.get('use_private_gitlab_repo'):
                spec['ckanContainerSpec']['imagePullSecrets'] = [{'name': 'private-gitlab-registry'}]
        elif create_type == 'from-gcloud-envvars':
            print(f'Creating Deis CKAN instance {instance_id} from gcloud envvars import')
            instance_env_yaml, image, solr_config, storage_path, instance_id = args[1:]
            db_migration_name = kwargs.get('db_migration_name')
            assert db_migration_name, 'creating from gcloud envvars without a db migration is not supported yet'
            if type(instance_env_yaml) == str:
                logs.info(f'Creating {instance_id}-envvars secret from file: {instance_env_yaml}')
                subprocess.check_call(
                    f'kubectl -n ckan-cloud create secret generic {instance_id}-envvars --from-file=envvars.yaml={instance_env_yaml}',
                    shell=True
                )
            else:
                logs.info(f'Creating {instance_id}-envvars secret from inline string')
                kubectl.update_secret(f'{instance_id}-envvars', {'envvars.yaml': yaml.dump(instance_env_yaml, default_flow_style=False)})
            spec = {
                'ckanPodSpec': {},
                'ckanContainerSpec': {'image': image},
                'envvars': {'fromSecret': f'{instance_id}-envvars'},
                'solrCloudCollection': {
                    'name': instance_id,
                    'configName': solr_config
                },
                'db': {
                    'name': instance_id,
                    'fromDbMigration':db_migration_name,
                    **({'dbPrefix': db_prefix} if db_prefix else {})
                },
                'datastore': {
                    'name': f'{instance_id}-datastore',
                    'fromDbMigration': db_migration_name,
                    **({'dbPrefix': db_prefix} if db_prefix else {})
                },
                'storage': {
                    'path': storage_path
                }
            }
        else:
            raise NotImplementedError(f'invalid create type: {create_type}')
        instance_kind = ckan_manager.instance_kind()
        instance = {
            'apiVersion': f'stable.viderum.com/v1',
            'kind': instance_kind,
            'metadata': {
                'name': instance_id,
                'namespace': 'ckan-cloud',
                'finalizers': ['finalizer.stable.viderum.com']
            },
            'spec': spec
        }
        subprocess.run('kubectl apply -f -', input=yaml.dump(instance).encode(), shell=True, check=True)
        return cls(instance_id, values=instance)

    def set_subdomain_route(self, router_type, router_name, route_type, router_annotations):
        assert router_type in ['traefik-subdomain']
        self.annotations.json_annotate(f'router-{route_type}-{router_name}', router_annotations)
